import logging
import sys
import os
import inspect
import warnings

log_setup_once = False


def setup_logging(name=None, log_level=logging.INFO):
	"""Set up logging."""
	global log_setup_once
	if not log_setup_once:
		log_setup_once = True
		logging.basicConfig(level=log_level)
		fmt = "%(asctime)s %(levelname)s (%(threadName)s) [%(name)s:%(lineno)s::%(funcName)s()] - %(message)s"
		colorfmt = "%(log_color)s{}%(reset)s".format(fmt)
		datefmt = "%Y-%m-%d %H:%M:%S"

		# Suppress overly verbose output that isn't helpful from some libraries we depend on
		for key in ["requests", "tensorboard", "urllib3", "aiohttp.access", "uamqp", "sqlalchemy", "sqlalchemy.engine.base", "matplotlib.font_manager"]:
			logging.getLogger(key).setLevel(logging.WARNING)

		# Enable debug logging for some insteresting libraries (for development)
		logging.getLogger("fk").setLevel(logging.DEBUG)

		try:
			from colorlog import ColoredFormatter

			logging.getLogger().handlers[0].setFormatter(ColoredFormatter(colorfmt, datefmt=datefmt, reset=True, log_colors={"DEBUG": "cyan", "INFO": "green", "WARNING": "yellow", "ERROR": "red", "CRITICAL": "red"}))
		except ImportError:
			# Ignore failures in loading color coded logs
			pass

		# Set this log level as default
		logger = logging.getLogger("")
		logger.setLevel(log_level)
	if not name:
		name = inspect.stack()[1][1]
	logger = logging.getLogger(name)
	try:
		import colored_traceback.auto
	except ImportError:
		pass

	return logger



import logging
import octomy.db
import os
import pprint
import pytest


logger = logging.getLogger(__name__)


here = os.path.dirname(__file__)
data_dir = os.path.join(here, "data")


# fmt: off
dummy_config= {
	"db-hostname":"hello.com"
	,"db-port":"1234"
	,"db-username":"arnold"
	,"db-password":"secret123"
	,"db-database":"mydb"
}
# fmt: on


def _test_db_get_same_twice():
	dbc1, err = octomy.db.get_database(config)
	dbc2, err = octomy.db.get_database(config)
	assert dbc1 == dbc2


@pytest.mark.asyncio
async def _test_get_database():
	config = {}
	dbc, err = octomy.db.get_database(config)
	assert(not dbc)
	assert(err)
	logger.warning(err)
	
	config = {
		'db-username': "test"
	}
	dbc, err = octomy.db.get_database(config)
	assert(not dbc)
	assert(err)
	logger.info(err)
	
	config = {
		'db-username': 1234
	}
	dbc, err = octomy.db.get_database(config)
	assert(not dbc)
	assert(err)
	logger.info(err)


	config = {
		  'db-username': "test"
		, 'db-port': "test"
	}
	dbc, err = octomy.db.get_database(config)
	assert(not dbc)
	assert(err)
	logger.info(err)
	
	config = {
		  'db-hostname': "test"
		, 'db-username': "test"
	}
	dbc, err = octomy.db.get_database(config)
	assert(not dbc)
	assert(err)
	logger.info(err)
	
	config = {
		  'db-hostname': "test"
		, 'db-username': "test"
		, 'db-password': "test"
	}
	dbc, err = octomy.db.get_database(config)
	assert(not dbc)
	assert(err)
	logger.info(err)
	
	config = {
		  'db-hostname': "test"
		, 'db-username': "test"
		, 'db-password': "test"
		, 'db-database': "test"
	}
	dbc, err = octomy.db.get_database(config, do_online=False)
	assert dbc, err
	assert not err, err
	logger.info(err)

	config = {
		  'db-hostname': "test"
		, 'db-username': "test"
		, 'db-password': "test"
		, 'db-database': "test"
	}
	dbc, err = octomy.db.get_database(config, do_online=True)
	assert not dbc, "Should fail"
	assert err, "Should fail"
	logger.info(err)




@pytest.mark.asyncio
async def _test_db():
	logger.info(f"here:{here}, data_dir:{data_dir}")
	config = {
		  'db-hostname': os.environ.get("TEST_DB_HOSTNAME")
		, 'db-port':     os.environ.get("TEST_DB_PORT")
		, 'db-database': os.environ.get("TEST_DB_DATABASE")
		, 'db-username': os.environ.get("TEST_DB_USERNAME")
		, 'db-password': os.environ.get("TEST_DB_PASSWORD")
	}
	db, err = octomy.db.get_database(config, )
	assert db, err
	# await db.preload_queries(f"{data_dir}/sql/")
	my_uuid = str(uuid.uuid4())
	# Assuming 'access.users.upsert' is a valid key and 'user_data' is a dictionary with the necessary parameters
	output=dict()
	matrix = [
		{
			  "query_key":"access.users.upsert"
			, "mode":"none"
			, "expected_result":None
			, "expected_error": 'query parameter missing: data, email, enabled'
			, "params": {'id': 1, 'name': 'John Doe'}
		}
		, {
			  "query_key":"access.users.upsert"
			, "mode": "one"
			, "expected_result": None
			, "expected_error": None
			, 'output_key': "my_id"
			, "params": {'id': 1, 'name': 'John Doe', 'data': 'balls', 'email': 'lolbob@sulla.com', 'enabled': True}
		}
	]
	for case in matrix:
		query_key = case.get("query_key")
		expected_result = case.get("expected_result")
		expected_error = case.get("expected_error")
		params = case.get("params")
		mode = case.get("mode", "none")
		logger.info(f"CASE query_key={query_key} expected_result={expected_result} expected_error={expected_error} params={params} mode={mode}")
		if mode == "none":
			actual_result, actual_error = await db.query_none(query_key = query_key, params=params)
		elif mode == "one":
			actual_result, actual_error = await db.query_one(query_key = query_key, params=params)
		elif mode == "many":
			actual_result, actual_error = await db.query_many(query_key = query_key, params=params)
		elif mode == "exists":
			actual_result, actual_error = await db.query_exists(query_key = query_key, params=params)
		else:
			raise Exception("incorrect mode. TODO: handle in testing")
		logger.info("EXPECTED RESULT:")
		logger.info(pprint.pformat(expected_result))
		logger.info("ACTUAL RESULT:")
		logger.info(pprint.pformat(actual_result))
		logger.info("EXPECTED ERROR:")
		logger.info(pprint.pformat(expected_error))
		logger.info("ACTUAL ERROR:")
		logger.info(pprint.pformat(actual_error))
		assert(actual_result == expected_result)
		assert(actual_error == expected_error)


def _test_query_key_to_sql_path():
	query_key = "octomy.db.get_now"
	logger.info(f"             query_key:{query_key}")
	sql_path = octomy.db.query_key_to_sql_path(query_key = query_key, relative=False, do_debug=True)
	logger.info(f"              sql_path:{sql_path}")
	assert bool(sql_path)
	assert os.path.exists(sql_path)



@pytest.mark.asyncio
async def test_convert_params():
	import pydantic #import BaseModel, StringConstraints, HttpUrl, validator
	import typing #import List, Optional, Annotated, Any, Dict
	class TestType(pydantic.BaseModel):
		test_str:str = "test str"
		test_constrained: typing.Annotated[str, pydantic.StringConstraints(max_length=255)] | None = None
		test_int:int = 1337
		test_bool:bool = True
		test_url:pydantic.HttpUrl = pydantic.HttpUrl("https://example.com:8888/somepath/endpoint?query1=lol#fragment")
	params_type = TestType()
	params_dict = octomy.db.pydantic_to_params(params_type)
	logger.info(f"FROM pydantic: {pprint.pformat(params_type)}")
	logger.info(f"TO params: {pprint.pformat(params_dict)}")
	
@pytest.mark.asyncio
async def _test_sql():
	config = {
		  'db-hostname': "test"
		, 'db-username': "test"
		, 'db-password': "test"
		, 'db-database': "test"
	}
	dbc, err = octomy.db.get_database(config, do_online=False)
	assert dbc, err
	assert not err, err
	query_key = "octomy.db.get_now"
	query = await dbc.load_query(query_key=query_key, do_debug=True)
	logger.info(f"query:'{query}'")
	assert bool(query), f"No query found for {query_key}"



from contextlib import asynccontextmanager
from fastapi import FastAPI, APIRouter, Depends, Request, HTTPException, status
from fastapi.responses import JSONResponse
from octomy.web.context import Context, get_context, context_dependency
from typing import List, Optional, Annotated
import jinja2.exceptions
import json
import logging
import pprint
import datetime
from fastapi.templating import Jinja2Templates


logger = logging.getLogger(__name__)


def prepare_templates():
	try:
		context = get_context()
		templates = Jinja2Templates(directory=f"{context.webroot}")
		templates.env.globals['datetime'] = datetime
		templates.env.globals['context'] = get_context()
		return  templates
	except Exception as e:
		logger.exception("No templates folder found, skipping...")
	return None

templates = prepare_templates()
 
def abortj(status_code, detail):
	content = {"status_code":status_code, "detail":detail}
	response = JSONResponse(status_code=status_code, content=content)
	logger.error("abortj")
	logger.warning(content)
	return response

json_content_type = "application/json"
html_content_type = "text/html"

FALLBACK_LIMIT = 100


def get_content_type(request):
	content_type = request.headers.get("content-type", None)
	if not content_type:
		return True
	if content_type == json_content_type:
		return "json"
	if content_type == html_content_type:
		return "html"
	else:
		raise HTTPException(
			status_code=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
			detail = f"Unsupported content type {content_type}")
	return False

def expect_json(request):
	content_type = request.headers.get("content-type", None)
	if not content_type:
		return True
	if content_type == json_content_type:
		return True
	else:
		raise HTTPException(
			status_code=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
			detail = f"Unsupported content type {content_type}")
	return False

def expect_html(request):
	content_type = request.headers.get("content-type", None)
	if not content_type:
		return True
	if content_type == html_content_type:
		return True
	else:
		raise HTTPException(
			status_code=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
			detail = f"Unsupported content type {content_type}")
	return False

async def db_get(request, context, query_key, params = dict(), mode="none", item_type = dict | None, prepare=True, do_debug=False):
	if do_debug:
		logger.info(f"db_get(request={request}, context={context}, query_key={query_key}, item_type={item_type})")
		logger.warning(f"PARAMS: {params}")
	params["limit"] = context.config.get("app-db-fetch-limit", FALLBACK_LIMIT)
	db_items, db_err = await context.db.key_query(query_key=query_key, params=params, mode=mode, item_type=item_type, prepare=prepare)
	#if None == db_items and mode != "none":
	#	raise Exception(db_err or f"No items found for '{query_key}'")
	return db_items, db_err

async def db_json(request, context, query_key, params = dict(), mode="none", item_type = dict | None, prepare=True, do_debug=False):
	if do_debug:
		logger.info(f"db_json(request={request}, context={context}, query_key={query_key}, item_type={item_type})")
	expect_json(request)
	return await db_get(request = request, context = context, query_key = query_key, params = params, mode=mode, item_type = item_type, prepare=prepare, do_debug = do_debug)


def handled_template(template_path, template_params):
	try:
		return templates.TemplateResponse(template_path, template_params)
	except Exception as e:
		if isinstance(e, jinja2.exceptions.TemplateError):
			logger.error("")
			logger.error(f"Template error: {type(e).__name__}: {e.message}")
			logger.error("")
			logger.warning(f"  for {template_path}")
			logger.warning("")
			logger.warning(f" with {pprint.pformat(template_params)}")
			logger.warning("")
			raise HTTPException(status_code = status.HTTP_500_INTERNAL_SERVER_ERROR, detail = "Template error")
		else:
			raise e




import logging
import octomy.utils
import os
import pprint
import pytest
import datetime


logger = logging.getLogger(__name__)


@pytest.mark.asyncio
async def _test_termcols():
	for name, color in vars(octomy.utils.termcol).items():
		if not name.startswith('__') and isinstance(color, str):
			bare = color.replace(octomy.utils.colbase, '')
			logger.info(f"{color}{name}\033[0m  {bare}")



from enum import Enum
from pydantic import BaseModel, Field
from typing import List, Optional, Union


MIMETYPE_OPENSCAD_MODEL:str = "model/openscad"
MIMETYPE_NTOP_MODEL:str = "model/ntop"
MIMETYPE_STL_MODEL:str = "model/stl"
MIMETYPE_PNG_IMAGE:str = "image/png"
MIMETYPE_JSON_DATA:str = "application/json"
MIMETYPE_TEXT:str = "text/plain"


class ParameterTypeEnum(str, Enum):
	NUMBER = 'NUMBER'
	STRING = 'STRING'
	BOOL = 'BOOL'
 
class NumericGeneratorParameter(BaseModel):
	name:str = "Numeric parameter name"
	value:Union[int, float]
 
class StringGeneratorParameter(BaseModel):
	name:str = "String parameter name"
	value:str
	
class BoolGeneratorParameter(BaseModel):
	name:str = "Bool parameter name"
	value:bool
 
class GeneratorParameters(BaseModel):
	action: str
	parameters: List[Union[NumericGeneratorParameter, StringGeneratorParameter, BoolGeneratorParameter]]
 
class NumericRangeGeneratorParameterSchema(BaseModel):
	name:str = "Numeric parameter name"
	min: Optional[Union[int, float]]
	step: Optional[Union[int, float]]
	max: Optional[Union[int, float]]
	default: Optional[Union[int, float]]
	unit: Optional[str]
	description:Optional[str] = "Parameter description"
	section:Optional[str] = "Parameter section"
	type:ParameterTypeEnum = ParameterTypeEnum.NUMBER
 
 
class NumericLabelGeneratorParameterValue(BaseModel):
	name:str = "Numeric parameter label"
	value:Union[int, float]
	
class NumericLabelGeneratorParameterSchema(BaseModel):
	name:str = "Numeric parameter name"
	labels: Optional[List[NumericLabelGeneratorParameterValue]]
	default: Optional[Union[int, float]]
	unit: Optional[str]
	description:Optional[str] = "Parameter description"
	section:Optional[str] = "Parameter section"
	type:ParameterTypeEnum = ParameterTypeEnum.NUMBER
 
class StringLabelGeneratorParameterValue(BaseModel):
	name:str = "String parameter label"
	value:str
 
class StringLabelGeneratorParameterSchema(BaseModel):
	name:str = "String parameter name"
	labels: Optional[List[StringLabelGeneratorParameterValue]]
	default:Optional[str]
	description:Optional[str] = "Parameter description"
	section:Optional[str] = "Parameter section"
	type:ParameterTypeEnum = ParameterTypeEnum.STRING
 
class StringLiteralGeneratorParameterSchema(BaseModel):
	name:str = "String parameter name"
	default:Optional[str]
	description:Optional[str] = "Parameter description"
	section:Optional[str] = "Parameter section"
	type:ParameterTypeEnum = ParameterTypeEnum.STRING
 
class BoolGeneratorParameterSchema(BaseModel):
	name:str = "Boolean parameter name"
	default:bool = False
	description:Optional[str] = "Parameter description"
	section:Optional[str] = "Parameter section"
	type:ParameterTypeEnum = ParameterTypeEnum.BOOL
 
 
class GeneratorTypeEnum(str, Enum):
	openscad = 'openscad'
	ntop = 'ntop'
	unknown = 'unknown'
 
 
class GeneratorMeta(BaseModel):
	generator_type:GeneratorTypeEnum = GeneratorTypeEnum.unknown
	version:Optional[str] = None
	class Config:  
		use_enum_values = True  # Use strings to represent enums
 
class GeneratorParametersSchema(BaseModel):
	generator_meta:GeneratorMeta
	parameters:List[Union[
		NumericRangeGeneratorParameterSchema
		, NumericLabelGeneratorParameterSchema
		, StringLiteralGeneratorParameterSchema
		, StringLabelGeneratorParameterSchema
		, BoolGeneratorParameterSchema] ]
 
class GeneratorInput(BaseModel):
	id: str
	parameters:GeneratorParameters
	result_url:Optional[str] = None
 
class GeneratorOutput(BaseModel):
	id: str
	error: Optional[Union[str, dict]]
	parameters: GeneratorParameters
	model_url:Optional[str] = None
	thumbnail_url:Optional[str] = None
	log_url:Optional[str] = None



import datetime
import math
from time import sleep

from octomy.utils import human_delta

import logging

logger = logging.getLogger(__name__)


class OnTheClockTimer:
	"""
	Class to provide a schedule from time of day + recurring interval as starting point
	Also provides convenience methods to make it easy to keep track of this schedule
	"""

	def __init__(self, start_time: datetime.time, interval: datetime.timedelta):
		self.start_time = start_time
		self.interval = interval

	def closest_start_time(self, now: datetime.datetime = datetime.datetime.now()) -> datetime.datetime:
		time_of_day = now.time()
		common_date = datetime.date(2019, 1, 1)
		p1 = datetime.datetime.combine(common_date, time_of_day)
		p2 = datetime.datetime.combine(common_date, self.start_time)
		from_start = p1 - p2
		interval_count = from_start / self.interval
		next_time = (math.floor(interval_count) + 1) * self.interval
		# logger.info(f"DATE: {now.date()}, TIME:{now.time()}, from_start:{from_start}, interval_count:{interval_count}, last_time={last_time}, next_time={next_time}")
		return datetime.datetime.combine(now.date(), self.start_time) + next_time

	def time_left(self, now: datetime.datetime = datetime.datetime.now()) -> datetime.timedelta:
		return self.closest_start_time(now=now) - now

	def wait_for_trigger(self, now: datetime.datetime = datetime.datetime.now()):
		iv = self.time_left(now=now)
		sec = iv.total_seconds()
		if sec > 0:
			logger.info(f"Waiting for {human_delta(iv)}")
			sleep(sec)
		return True

	def __str__(self):
		return f"OnTheClockTimer(start_time={self.start_time}, interval={self.interval})"



import logging
import base64

oktopus_folder_name = '.oktopus'
oktopus_version = '0.0.15'

logger = logging.getLogger(__name__)

def file_properties(id, key):
	properties = {
		  'oktopus_source_id': id
		, 'oktopus_parameter_key': key
		, 'oktopus_version': oktopus_version
	}
	return properties


from octomy.cad.generators.ntop import nTopGenerator
from octomy.cad.generators.openscad import OpenSCADGenerator
from octomy.cad.types import *
from octomy.storage.google_drive import GoogleDrive, FOLDER_MIMETYPE, folder_id as google_drive_folder_id


class GeneratorTool:
	def __init__(self, storage, oktopus_folder_name):
		self.oktopus_folder_id = None
		self.storage = storage
		self.oktopus_folder_name = oktopus_folder_name
		self.ensure_oktopus_folder()

	def find_oktopus_folder(self):
		self.oktopus_folder_id = None
		q = f"'{google_drive_folder_id}' in parents and name = '{self.oktopus_folder_name}' and trashed = false and mimeType = '{FOLDER_MIMETYPE}'"
		folders = self.storage.search_files(q = q)
		if folders and len(folders) >= 0:
			self.oktopus_folder_id = folders[0].get("id")
		return self.oktopus_folder_id

	def ensure_oktopus_folder(self):
		#logger.info("ENSURE DERIVATIVES")
		self.oktopus_folder_id = self.find_oktopus_folder()
		if not self.oktopus_folder_id:
			ret = self.storage.create_folder(file_metadata={"name":self.oktopus_folder_name, "parents":[google_drive_folder_id]})
			if ret:
				#logger.info(f"ret:{pprint.pformat(ret)}")
				self.oktopus_folder_id = ret
		return self.oktopus_folder_id

	def find_source_folder(self, name):
		q = f"'{self.oktopus_folder_id}' in parents and name = '{name}' and trashed = false and mimeType = '{FOLDER_MIMETYPE}'"
		source_folder_id = None
		folders = self.storage.search_files(q = q)
		if folders and len(folders) >= 0:
			source_folder_id = folders[0].get("id")
		return source_folder_id
		
	def ensure_source_folder(self, name):
		source_folder_id = self.find_source_folder(name)
		if not source_folder_id:
			ret = self.storage.create_folder(file_metadata={"name":name, "parents":[self.oktopus_folder_id]})
			if ret:
				#logger.info(f"ret:{pprint.pformat(ret)}")
				source_folder_id = ret
		return source_folder_id
	
	def find_part_folder(self, source_folder_id, name):
		q = f"'{source_folder_id}' in parents and name = '{name}' and trashed = false and mimeType = '{FOLDER_MIMETYPE}'"
		part_folder_id = None
		folders = self.storage.search_files(q = q)
		if folders and len(folders) >= 0:
			part_folder_id = folders[0].get("id")
		return part_folder_id
		
	def ensure_part_folder(self, source_folder_id, name):
		part_folder_id = self.find_part_folder(source_folder_id, name)
		if not part_folder_id:
			ret = self.storage.create_folder(file_metadata={"name":name, "parents":[source_folder_id]})
			if ret:
				#logger.info(f"ret:{pprint.pformat(ret)}")
				part_folder_id = ret
		return part_folder_id



	def upload_derivative(self, local_path, parent_id, name, mimetype, properties = dict(), do_debug = False):
		if do_debug:
			logger.info(f"Uploading file '{local_path}'")
			logger.info(f"  of type '{mimetype}'")
			logger.info(f"  with properties '{pprint.pformat(properties)}'")
			logger.info(f"  to drive at '{parent_id}/{name}'")
		file_metadata = {"parents":[parent_id], "name": name }
		if properties:
			file_metadata["properties"] = properties
		file_id = self.storage.upload_file(
			  file_path = local_path
			, file_metadata = file_metadata
			, mimetype = mimetype)
		#logger.info("UPLOADED '{local_path}' to '{name}' gave ID '{file_id}'")
		return file_id

	def set_drive_thumbnail_from_data(self, file_id, data, do_debug = True):
		if do_debug:
			logger.info(f"Setting thumbnail of {file_id} to raw data ({len(data)} bytes)")
		if not file_id:
			logger.warning("No id")
			return
		metadata = {
			"contentHints": {
				"thumbnail": {
					"image": base64.urlsafe_b64encode(data).decode('utf8'),
					"mimeType": MIMETYPE_PNG_IMAGE,
				}
			}
		}
		res = self.storage.update_file_metadata(file_id, metadata)


	def set_drive_thumbnail_from_local_path(self, file_id, thumbnail_path, do_debug = True):
		if do_debug:
			logger.info(f"Setting thumbnail of {file_id} to content of '{thumbnail_path}")
		if not file_id:
			logger.warning("No id")
			return
		with open(thumbnail_path, "rb") as f:
			return self.set_drive_thumbnail_from_data(file_id, f.read(), do_debug = do_debug)


storage = GoogleDrive()
common = GeneratorTool(storage = storage, oktopus_folder_name = google_drive_folder_id)

generators_map = {
	GeneratorTypeEnum.openscad: OpenSCADGenerator(common = common)
	, GeneratorTypeEnum.ntop: nTopGenerator(common = common)
}



from fastapi import Request, HTTPException
from fastapi.responses import HTMLResponse
from pathlib import Path
import datetime
import frontmatter
import jinja2
import logging
import markdown
import os
import pprint
import traceback


logger = logging.getLogger(__name__)


def gen_tags(parts):
	tags = ["-".join(parts[:-1])] + [part for part in parts]
	# logger.info(f"TAGS: {pprint.pformat(parts)} --> {tags}")
	return  tags

def url_for(index):
	return "http://TODO_MAKE_URLFOR/"+str(index)

class AutoRouter:

	exts = ['html', 'md']
	#md_extensions = ['tables', 'md_in_html', 'footnotes', 'fenced_code', 'def_list', 'attr_list', 'abbr', 'toc', 'smarty']
	md_extensions = ['tables', 'md_in_html', 'footnotes', 'fenced_code', 'def_list', 'abbr', 'toc', ]

	def __init__(self, app, context):
		self.app = app
		self.context = context
		self.env = self.prepare_env()
		self.auto_routes = {}
		self.auto_route_all(self.context.webroot)
		#logger.info(f"ROUTES: {pprint.PrettyPrinter(indent=4, compact=False).pformat(self.auto_routes)}")


	def prepare_env(self):
		loader = jinja2.ChoiceLoader(
			[
				jinja2.FileSystemLoader(self.context.webroot)
	#			, jinja2.FileSystemLoader(os.path.join(project_folder, os.path.dirname(template_name)))
			]
		)
		env = jinja2.Environment(loader=loader)
		env.globals['datetime'] = datetime
		env.globals['context'] = self.context
#		env.globals['current_user'] = user
#		env.globals['current_user_can'] = current_user_can
#		env.globals['url_for'] = url_for
		#env.globals.update(get_expiry=octomy.utils.expiry_cache.get_expiry)
		#env.filters['myfilter'] = filter
		return env


	def _subpages(self, path, max_level=2):
		out = {}
		#space = max_level * "   "
		parts = os.path.normpath(path).strip(os.path.sep).split(os.path.sep) 
		path = "/".join(parts)
		obj = self.auto_routes.get(path, dict())
		if max_level > 1:
			out[path] = obj
			if obj:
				subs = obj.get("children", list())
				for sub in subs:
					subsubs = self._subpages(sub, max_level - 1)
					for subsub in subsubs:
						if subsub not in out:
							subobj = self.auto_routes.get(subsub, dict())
							if subobj:
								out[subsub] = subobj
		return out



	def _perform_markdown(self, template_string):
		template_string = markdown.markdown(template_string, extensions = self.md_extensions)
		return template_string
	
	def _perform_frontmatter(self, template_string):
		page = frontmatter.loads(template_string)
		logger.info(f"FRONTMATTER:{dict(page)}")
		return page.content, dict(page)
		

	def _perform_jinja_string(self, template_string, data=dict()):
		template = self.env.from_string(template_string)
		if not template:
			return None, f"Error creating template for '{template_fn}'"
		#logger.info(f"template_fn:{template_fn}")
		html = template.render(title="TODO: Set title from context", **data)
		return html, None

	def _perform_jinja_file(self, template_fn, data=dict()):
		template_string = ""
		if not os.path.isabs(template_fn):
			template_fn = os.path.join(self.context.webroot, template_fn)
		# logger.info(f"AMENDED PATH:{template_fn}")
		with open(template_fn, "r") as template_file:
			template_string = template_file.read()
		return self._perform_jinja_string(template_string, data=data)

	def _perform_wrapping(self, template_string:str, data:dict = dict(), pre_wrappers:list = list(), post_wrappers:list = list()):
		out = ""
		for wrapper in pre_wrappers:
			content, error = self._perform_jinja_file(template_fn = wrapper, data = data)
			if error:
				logger.warning(f"Error appending pre wrapper '{wrapper}', skipping: {error}")
			else:
				out += content
		out += template_string
		for wrapper in post_wrappers:
			content, error = self._perform_jinja_file(template_fn = wrapper, data = data)
			if error:
				logger.warning(f"Error appending post wrapper '{wrapper}', skipping: {error}")
			else:
				out += content
		return out

	def _find_index(self, path):
		template_fn = None
		type = None
		for ext in self.exts:
			candidate = f"{self.context.webroot}/{path}/index.{ext}"
			if os.path.isfile(candidate):
				template_fn = candidate;
				type = ext
				break
		return template_fn, type

	def _get_path_parts(self, path):
		parts = os.path.normpath(path).strip(os.path.sep).split(os.path.sep) 
		return parts

	def handler_raw(self, path, request):
		try:
			template_fn, type = self._find_index(path)
			if not template_fn:
				return None, f"No index file found for {path}"
			template_string = ""
			parts = self._get_path_parts(path)
			subs = self._subpages(path, 4)
			logger.info(f"Reading template {template_fn}:")
			with open(template_fn, "r") as template_file:
				template_string = template_file.read()
			template_string, frontmatter = self._perform_frontmatter(template_string)
			if 'md' == type:
				template_string = self._perform_markdown(template_string)
			data = {
				  'type': type
				, 'subpages': subs
				, 'request': request
				, 'frontmatter': frontmatter
				, 'context': self.context
			}
			if frontmatter:
				pre_wrappers = frontmatter.get('pre_wrappers', list())
				post_wrappers = frontmatter.get('post_wrappers', list())
				template_string = self._perform_wrapping(template_string, data = data, pre_wrappers = pre_wrappers, post_wrappers = post_wrappers)
			html, err = self._perform_jinja_string(template_string, data)
			return html, err
		except Exception as e:
			logger.exception(f"Could not handle {path}")
			return None, f"Error rendering template '{template_fn}' with error\n{traceback.format_exc()}"


	def auto_route_all(self, base):
		for ext in self.exts:
			for path in Path(base).rglob(f'index.{ext}'):
				self.auto_route_path(path)
	

	def auto_route_path(self, template_fn):
		dir = Path(os.path.relpath(os.path.dirname(template_fn), self.context.webroot))
		self.auto_routes[str(dir)] = { "children":[], "path": str(dir), "title": os.path.basename(str(dir)).capitalize() }
		name = os.path.basename(dir)
		#logger.info(f"GOT ROUTE{pprint.pformat(dir.parts)} {name}")
		self.app.add_api_route(path=f"/{dir}", endpoint=self.handler, methods=["GET"], tags=gen_tags(dir.parts), name=name, response_class=HTMLResponse)
		acc = []
		for part in dir.parts:
			key = "/".join(acc)
			node = self.auto_routes.get(key, dict())
			children = node.get("children", list())
			childkey = f"{key}/{part}"
			if childkey not in children:
				children.append(childkey)
			node["children"] = children
			node["path"] = str(key)
			node["title"] = os.path.basename(str(key)).capitalize()
			self.auto_routes[key] = node
			#index_page(str(key))
			acc.append(part)
		return {"dir":dir, "name":name, }
	

	def _generr(self, err):
		return f"""
		<h1>An error occurred</h1>
		<pre>
		{err}
		</pre>
		"""
	def handler(self, request:Request):
		path = request.scope['route'].path
		logger.info(f"GOT HANDLER{pprint.pformat(path)}, {pprint.pformat(request.path_params)} {request.url.path}")
		res, err = self.handler_raw(path, request)
		if err:
			#return HTTPException(status_code=500, detail=err)
			return HTMLResponse(content=self._generr(err), status_code=200)
		return HTMLResponse(content=res, status_code=200)



 #!/bin/env python3
from octomy import utils
from octomy.cad.generators import generators_map, common, file_properties, oktopus_folder_name
from octomy.cad.types import *
from octomy.storage.google_drive import GoogleDrive, FOLDER_MIMETYPE, folder_id as google_drive_folder_id
from enum import Enum
from pydantic import BaseModel, Field
from typing import List, Optional, Union
import base64
import datetime
import jinja2
import logging
import math
import os
import pathlib
import pprint
import random
import re
import time


logger = logging.getLogger(__name__)


DEFAULT_UNIT = "mm"

MIN_UPDATE_FREQ = 1000 * 60 # Update at least every minute

# Dir where data is kept
data_dir = "data"

# Maximum filesize allowed
data_file_max_size = 1024*1024*1024*2 # 2 GiB



# Number of floats per record (bar) of data
data_float_bytes = 4
#open: float
#high: float
#low: float
#close: float
#volume: float
#trade_count: Optional[float]
#vwap: Optional[float]
#pad: float #8 x float alignment padding

# Number of bytes per float
data_columns = 8

# Maximum number onf indices per file
data_file_max_index = data_file_max_size / (data_columns * data_float_bytes)


parts_by_id = dict()


model_format_map = {
	  MIMETYPE_OPENSCAD_MODEL: "OpenSCAD"
	, MIMETYPE_NTOP_MODEL: "nTop"
}

def understand_model_format(mimetype):
	if not mimetype:
		return False, "Unknown"


def millinow():
	tz = datetime.timezone.utc
	now = datetime.datetime.now(tz)
	epoch = datetime.datetime(1970, 1, 1, tzinfo = tz) # use POSIX epoch
	posix_timestamp_millis = (now - epoch) // datetime.timedelta(milliseconds = 1) 
	return posix_timestamp_millis



##################################################################


def calculate_filenames(id, parameters, model_type="scad", output_type="stl"):
	# Add millisecond timestamp to avoid problems in the (rare) case where two calls are made for the same file at the same time
	now = millinow()
	local_name_base = f"/tmp/{id}_{now}/part"
	local_name_input = f"{local_name_base}.{model_type}"
	local_name_output = f"{local_name_base}.{output_type}"
	local_name_meta = f"{local_name_base}.json"
	local_name_log = f"{local_name_base}.log"
	local_name_thumbnail = f"{local_name_base}.png"
	key = utils.calculate_unique_key(parameters.json())
	storage_name_base = f"{id}_{key}"
	storage_name_output = f"{storage_name_base}.{output_type}"
	storage_name_meta = f"{storage_name_base}.json"
	storage_name_log = f"{storage_name_base}.log"
	storage_name_thumbnail = f"{storage_name_base}.png"
	return {
		  "key": key
		, "now": now
		, "local_name_base": local_name_base
		, "local_name_input": local_name_input
		, "local_name_output": local_name_output
		, "local_name_meta": local_name_meta
		, "local_name_log": local_name_log
		, "local_name_thumbnail": local_name_thumbnail
		, "storage_name_base": storage_name_base
		, "storage_name_output": storage_name_output
		, "storage_name_meta": storage_name_meta
		, "storage_name_log": storage_name_log
		, "storage_name_thumbnail": storage_name_thumbnail
	}

def prepare_gen_job(part_id, part_folder_id):
	return {
		  "part_id": part_id
		, "parameters": parameters
		, "part_folder_id": part_folder_id
		, "storage_name_base": storage_name_base
		, "storage_name_output": storage_name_output
		, "storage_name_meta": storage_name_meta
		, "storage_name_log": storage_name_log
		, "storage_name_thumbnail": storage_name_thumbnail
	}

def determine_file_extension(file_path: str):
	if not file_path:
		return None, "No file path"
	index = file_path.rfind(".")
	if index < 0:
		return None, "No file extention separator"
	file_ending = file_path[index + 1 :]
	if not file_ending:
		return None, "No file extention"
	return file_ending, None

def file_extension_is_binary(ext):
	return not ext in ["md", "htm", "html", "inc", "txt", "svg", "csv", "json", "jinja2", "jinja", "j2", "tpl", "scad"]

def file_is_binary(path):
	ext, err = determine_file_extension(path)
	if ext:
		return file_extension_is_binary(ext)
	# Fall back to binary to be safe
	return True


def properties_to_query(properties):
	if not properties:
		return ""
	all = list()
	for key, value in properties.items():
		all.append(f"properties has {{ key='{key}' and value='{value}' }}")
	return "and ".join(all)

def pretty_name(basename):
	return os.path.splitext(basename)[0];

##################################################################


def convert_schema(schemas, version = None, do_debug = False):
	output_schema = list()
	for input_parameter in schemas:
		output_parameter = None
		if do_debug:
			logger.info(f"input_parameter:\n{pprint.pformat(input_parameter)}")
		default = input_parameter.get("default", dict())
		if not default:
			logger.warning(f"No default for {pprint.pformat(input_parameter)}, skipping parameter")
			continue
		_type = default.get("type")
		if not _type:
			logger.warning(f"No type for {pprint.pformat(input_parameter)}, skipping parameter")
			continue
		section = input_parameter.get("section")
		if not section:
			if do_debug:
				logger.warning(f"No section for {pprint.pformat(input_parameter)}, skipping parameter")
			continue
		schema = input_parameter.get("schema", dict())
		name = input_parameter.get("name")
		default_unit = input_parameter.get("unit")
		default_value = default.get("value")
		description = input_parameter.get("description")
		if _type == "number":
			if schema:
				if schema.get("labels"):
					if do_debug:
						logger.info(f"Adding numeric labels parameter {name} ({pprint.pformat(schema)})")
					labels = list()
					for label_value, label_name in schema.get("labels").items():
						if do_debug:
							logger.info(f"  + {label_name}={label_value}({type(label_value)})")
						labels.append(NumericLabelGeneratorParameterValue(name=str(label_name), value=label_value))
					output_parameter = NumericLabelGeneratorParameterSchema(
						  name = name
						, labels = labels
						, default = default_value
						, unit = default_unit
						, description = description
						, section = section)
				elif schema.get("values"):
					if do_debug:
						logger.info(f"Adding numeric values parameter {name} ({pprint.pformat(schema)})")
					labels = list()
					for label_value in schema.get("values"):
						#logger.info(f"  + value={label_value}")
						labels.append(NumericLabelGeneratorParameterValue(name=str(label_value), value=label_value))
					output_parameter = NumericLabelGeneratorParameterSchema(
						  name = name
						, labels = labels
						, default = default_value
						, unit = default_unit
						, description = description
						, section = section)
				else:
					#logger.info(f"Adding numeric range parameter {name} ({pprint.pformat(schema)})")
					output_parameter = NumericRangeGeneratorParameterSchema(
						  name = name
						, min = schema.get("min")
						, step = schema.get("step")
						, max = schema.get("max")
						, default = default_value
						, unit = default_unit
						, description = description
						, section = section)
		elif _type == "string":
			if schema.get("labels"):
				if do_debug:
					logger.info(f"Adding string labels parameter '{name}' ({pprint.pformat(schema)})")
				labels = list()
				for label_value, label_name in schema.get("labels").items():
					logger.info(f"  + {label_name}={label_value}")
					labels.append(StringLabelGeneratorParameterValue(name=label_name, value=label_value))
				output_parameter = StringLabelGeneratorParameterSchema(
					  name = name
					, labels = labels
					, default = default_value
					, description = description
					, section = section)
			elif schema.get("values"):
				if do_debug:
					logger.info(f"Adding string values parameter '{name}' ({pprint.pformat(schema)})")
				labels = list()
				for label_value in schema.get("values"):
					logger.info(f"  + value={label_value}")
					labels.append(StringLabelGeneratorParameterValue(name=label_value, value=label_value))
				output_parameter = StringLabelGeneratorParameterSchema(
					  name = name
					, labels = labels
					, default = default_value
					, description = description
					, section = section)
			else:
				if do_debug:
					logger.info(f"Adding literal string parameter {name} ({pprint.pformat(schema)})")
				output_parameter = StringLiteralGeneratorParameterSchema(
					  name = name
					, default = default_value
					, description = description
					, section = section)
		elif _type == "bool":
			if do_debug:
				logger.info(f"Adding bool parameter {name} {pprint.pformat(schema)}")
			output_parameter = BoolGeneratorParameterSchema(
			  name = name
			, default = default_value
			, description = description
			, section = section)
		if output_parameter:
			#logger.info(f"Appending output parameter: {pprint.pformat(output_parameter)}")
			output_schema.append(output_parameter)
			output_parameter = None
		else:
			logger.warning(f"Could not generate output parameter from:\n{pprint.pformat(input_parameter)}")
	#logger.info("-------------------OUTPUT_schema")
	#logger.info(pprint.pformat(output_schema))
	#logger.info("-------------------OUTPUT_schema")
	generator_meta = GeneratorMeta(generator_type = GeneratorTypeEnum.openscad, version = version)
	schema = GeneratorParametersSchema(generator_meta = generator_meta, parameters = output_schema)
	#logger.info(pprint.pformat(json.loads(    schema.json()           )))
	return schema


##################################################################


class Parts:
	def __init__(self):
		utils.inspectify(self)
		self.oktopus_folder_id = None
		self.storage = GoogleDrive()
		self.common = common
		self.update_files()
		self.default_query_stump = "supportsAllDrives = true and includeItemsFromAllDrives = true and corpora = allDrives"

	def source_exists(self, id:str):
		return self.storage.file_exists(id)

	def source(self, id:str, do_debug = False):
		source = self.storage.download_file(id)
		if do_debug:
			logger.info(f"Got source for {id}: {pprint.pformat(source)}")
		if not source:
			return None, "No source for id"
		try:
			source = str(source, 'utf-8')
		except Exception as e:
			logger.error(f"error converting source to utf-8 string for id: {id}")
			logger.warning(f"The related source was {source}")
			logger.exception(e)
		return source, None

	def source_schema(self, id:str, do_debug = False):
		source, source_err = self.source(id)
		if not source:
			return None, source_err
		
		generator_type = GeneratorTypeEnum.openscad
		generator = generators_map.get(generator_type)
		if not generator:
			return None, f"Unknown or unsupported generator '{generator_type}' selected"
		#logger.info(f"Source was {pprint.pformat(source)}")
		schemas = generator.get_schema(source = source, do_debug = do_debug)
		#schemas = parse_customize(source, do_debug)
		if not schemas:
			return None, "No schema for source"
		#logger.info(f"Found schema for id '{id}': {pprint.pformat(schemas)}")
		version = generator.get_version()
		schema = convert_schema(schemas = schemas, version = version, do_debug = do_debug)
		return schema, None


	def find_complete_part_folder(self, source_name, part_name):
		source_folder_id = self.find_source_folder(source_name)
		part_folder_id = source_folder_id and self.find_part_folder(source_folder_id, part_name)
		return part_folder_id


	def _find_derivative(self, id:str, parameters:GeneratorParameters, mimetype = MIMETYPE_STL_MODEL, do_debug = False):
		if not parameters:
			logger.warning(f"No parameters ({parameters}) provided while finding derivative for '{id}' ({mimetype})")
			return None
		key = utils.calculate_unique_key(parameters.json())
		properties_string = properties_to_query(file_properties(id, key))
		q = f"{properties_string} and trashed = false and mimeType = '{mimetype}'"
		files = self.storage.search_files(q = q, fields=['properties'])
		# logger.info(f"Derivative search for id='{id}', key='{key}', mimetype='{mimetype}' resulted in {pprint.pformat(files)}")
		# File existed, return it's id
		if files and len(files) >= 0:
			ret_id = files[0].get("id")
			if do_debug:
				logger.info(f"Found existing derivative for '{id}' ({mimetype})")
			return ret_id
		return None


	def _get_default_parameters(self, schema, do_debug = False):
		# Use default paramters if noen given
		parameters = list()
		
		#logger.info(f"schema 8=======D ~~~  {pprint.pformat(schema)}")
		for parameter in schema.parameters:
			#logger.info(f" + {parameter.name}: {parameter.default}")
			#parameters.append( NumericGeneratorParameter( name = parameter.name, value = parameter.default))
			out = None
			if parameter:
				if isinstance(parameter, BoolGeneratorParameterSchema):
					out = BoolGeneratorParameter(name = parameter.name, value = parameter.default)
				elif isinstance(parameter, (NumericRangeGeneratorParameterSchema, NumericLabelGeneratorParameterSchema)):
					out = NumericGeneratorParameter(name = parameter.name, value = parameter.default)
				elif isinstance(parameter, (StringLiteralGeneratorParameterSchema, StringLabelGeneratorParameterSchema)):
					out = StringGeneratorParameter(name = parameter.name, value = parameter.default)
				if out:
					parameters.append(out)
			if do_debug:
				logger.info(f"{pprint.pformat(parameter)} --> {pprint.pformat(out)}")
		if do_debug:
			logger.info(f"Generated default parameters: {pprint.pformat(parameters)}")
		parameters = GeneratorParameters(action="generate", parameters = parameters)
		return parameters


	def _generate_worker(self, id:str, parameters:GeneratorParameters, do_stl=True, do_png=True, do_debug=False):
		if not id:
			return None, {"message":"No id provided while generating", "error": None}
		if not parameters:
			return None, {"message":"No parameters provided while generating {id}", "error": None}
		# We need the binary raw file so not using self.source()
		source = self.storage.download_file(id)
		if not source:
			return None, {"message":"Could not download source for {id}", "error": None}
		source_meta = self.storage.file_meta(id)
		if not source_meta:
			return None, {"message":"Could not download source for {id}", "error": None}
		logger.info(f"Source meta:{source_meta}")
		key = key = utils.calculate_unique_key(parameters.json())
		if not key:
			return None, {"message":"Could not generate derivative filenames for {id}", "error": None}
		source_name = source_meta.get("name", id)
		generator_type = GeneratorTypeEnum.openscad
		generator = generators_map.get(generator_type)
		if not generator:
			return None, f"Unknown or unsupported generator '{generator_type}' selected"
		return generator.generate(source_name = source_name, source = source, parameters = parameters, id = id, do_stl = do_stl, do_png = do_png, do_debug = do_debug)


	def generate(self, id:str, parameters:GeneratorParameters, do_stl = True, do_png = True, do_debug = False):
		stl_file_id = None
		png_file_id = None
		json_file_id = None
		log_file_id = None
		if do_stl:
			stl_file_id = self._find_derivative(id = id, parameters = parameters, mimetype = MIMETYPE_STL_MODEL)
		if do_png:
			png_file_id = self._find_derivative(id = id, parameters = parameters, mimetype = MIMETYPE_PNG_IMAGE)
		ret = None
		err = None
		if (do_stl and stl_file_id) or (do_png and png_file_id):
			# Model file found, gather the rest
			json_file_id = self._find_derivative(id = id, parameters = parameters, mimetype = MIMETYPE_JSON_DATA)
			log_file_id = self._find_derivative(id = id, parameters = parameters, mimetype = MIMETYPE_TEXT)
			ret = {
				  "model_id": stl_file_id
				, "parameters_id": json_file_id
				, "log_id": log_file_id
				, "thumbnail_id": png_file_id}
			err = None
			if do_debug:
				logger.info(f"Found existing derivative for '{id}'")
		else:
			# Model file did not exist, generate them
			logger.warning(f"No existing derivative found for '{id}', generating...")
			ret, err = self._generate_worker(id, parameters, do_stl=do_stl, do_png=do_png, do_debug=do_debug)
		if do_debug:
			logger.info(f"Returning {pprint.pformat(ret)} ({pprint.pformat(err)})")
		return ret, err


	def _generate_default_part_thumbnail_id(self, id, do_debug = True):
		if do_debug:
			logger.info(f"Generating default thumbnail for openscad file for {id}")
		source_schema, source_schema_err = self.source_schema(id = id, do_debug = do_debug)
		if not source_schema:
			logger.warning(f"Error fetching source_schema for id {id}: {pprint.pformat(source_schema_err)}")
			return None
		parameters = self._get_default_parameters(source_schema, do_debug = do_debug)
		# Perform a full generate while we are at it (it will take the same amount of time and will have to be done later anyways
		ret, err = self.generate(id = id, parameters = parameters, do_stl = True, do_png = True, do_debug = do_debug)
		if not ret or err:
			logger.warning(f"Error generating thumbnail: {pprint.pformat(err)}")
			return None
		thumbnail_id = ret.get("thumbnail_id")
		return thumbnail_id


	def _generate_default_part_thumbnail_url(self, id, do_debug = True):
		thumbnail_id = self._generate_default_part_thumbnail_id(id = id, do_debug = do_debug)
		if not thumbnail_id:
			logger.warning(f"No thumbnail found for id: {id}")
			return None, None
		data = self.get_derivative(id, thumbnail_id)
		if not data:
			logger.warning(f"Could not fetch data for thumbnail_id {pprint.pformat(thumbnail_id)}")
			return None, None
		if do_debug:
			logger.info(f"Setting default thumbnail for {id} to png of {len(data)} bytes")
		ret = self.set_drive_thumbnail_from_data(id, data, do_debug = do_debug)
		return ret, thumbnail_id


	def update_files(self, generate_thumbnails = False, do_debug = False):
		q = f"'{google_drive_folder_id}' in parents and name contains '.scad' and trashed=false"
		#self.files = files = self.storage.search_files(q, ["thumbnailLink"])

		files, folders = self.storage.recursive_search_worker(
			  folder_id = google_drive_folder_id
			#, f"trashed = false and name contains '.scad'"
			#, f"trashed = false and name not contains '.'"
			#, q_files = f"trashed = false and (name contains '.scad' or name contains '.ntop')" #TODO: Enable ntop support
			, q_files = f"trashed = false and (name contains '.scad')"
			, q_folders = f"trashed = false"
			, fields = ["thumbnailLink"]
			, path = list()
			, do_debug = do_debug) 
		
		self.files = files
		self.folders = folders
		
		#logger.info(pprint.pformat(self.files))
		#logger.info(f"PARTS:")
		self.parts_by_id = dict()
		self.parts = list()
		self.last_file_update = millinow()
		
		time_budget_millis = 60 * 1000 # a maximum of one minute of generation time is allowed per update
		update_start = millinow()
		for file in self.files:
			id = file.get("id")
			if id:
				mimetype = file.get("mimetype")
				format = understand_model_format(mimetype)
				if format:
					filename = file.get("name")
					path = file.get("path")
					tags = list()
					for p in path:
						tags.append(p.get("name"))
					basename = os.path.basename(filename)
					thumbnail_url = file.get("thumbnailLink")
					thumbnail_id = None
					if generate_thumbnails:
						logger.info(f"Doing thumbnail generation for {filename}")
						thumbnail_id = self._generate_default_part_thumbnail_id(id, do_debug = do_debug)
					else:
						source_schema, source_schema_err = self.source_schema(id = id, do_debug = do_debug)
						if source_schema:
							parameters = self._get_default_parameters(source_schema, do_debug = do_debug)
							if parameters:
								thumbnail_id = self._find_derivative(id = id, parameters = parameters, mimetype = MIMETYPE_PNG_IMAGE)
					part = {
						  "id":id
						, "title": basename
						, "mimetype": mimetype
						, "description": "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
						, "filename": filename
						, "pretty_name": pretty_name(basename)
						, "tags": tags
						, "thumbnail_id": thumbnail_id
					}
					self.parts.extend([part])
					self.parts_by_id[id] = part
					#logger.info(f"Registering part: {pprint.pformat(part)}")
				else:
					logger.warning(f"Skipping file of unknown type: {mimetype}")
			else:
				logger.warning(f"Skipping part without id: {pprint.pformat(part)}")
					
	def maybe_update_files(self):
		since = millinow() - self.last_file_update
		if since > MIN_UPDATE_FREQ:
			self.update_files()
			return True
		return False

	def get_meta(self, id):
		self.maybe_update_files()
		return self.parts_by_id.get(id)

	def get_derivative(self, id, derivative_id):
		# TODO: For security, we should arrive at the ID fo the file to download indirectly
		data = self.storage.download_file(derivative_id)
		return data

	def get_parts(self):
		self.maybe_update_files()
		return self.parts



#!/bin/env python3
import aiofiles
import asyncio
import logging
import os
import pathlib
import pprint
import psycopg
import psycopg_pool
import re


logger = logging.getLogger(__name__)


def stripes(cols=79, rows=3, on=3, off=3, on_char="#", off_char=" ", offset=0, sheer=1):
	out=""
	alphabet = on_char * on + off_char * off
	alphabet_len=len(alphabet)
	for row in range(0, rows):
		index = offset
		for col in range(0, cols):
			alphabet_index = index % alphabet_len
			out += alphabet[alphabet_index]
			index += 1
		offset+=sheer
		out+="\n"
	return out

def log_exception(e):
	if isinstance(e, asyncpg.PostgresError):
		# Log detailed database error information
		logger.error(f"+ Postgres error ({type(e).__name__}): e={e}")
		logger.error(f"+ Detail: {e.detail}")
		logger.error(f"+ Hint: {e.hint}")
	else:
		# Log non-database errors
		logger.error(f"+ NON DB ERROR ({type(e).__name__}): e={e}")

	# Log traceback
	logger.error("Traceback (most recent call last):")
	traceback.print_exception(type(e), e, e.__traceback__)
	
def log_db_exception(e, query_key, params):
	logger.error("\n" * 2)
	logger.error("#" * 50)
	logger.error(f"Querying: {query_key}")
	logger.error(f"Using data: {pprint.pformat(params)}")
	logger.error(f"Failed with:")
	log_exception(e)
	logger.error("#" * 45)
	logger.error("\n" * 2)

class Database:
	def __init__(self, db_uri):
		self.db_uri = db_uri
		self.query_dir = None
		self.queries = dict()
		self.prepared = dict()
		self.pool = None


	async def prepare_query(self):
		 statement = await conn.prepare("SELECT * FROM my_table WHERE id = $1")
		 
	async def preload_queries(self, path):
		"""Recursively load and register queries from .sql files in given path."""
		self.query_dir = pathlib.Path(path)
		for sql_file in self.query_dir.rglob('*.sql'):
			# some/path/file.sql --> some.path.file
			query_key = sql_file.relative_to(self.query_dir).with_suffix('').as_posix().replace('/', '.')
			with open(sql_file, 'r', encoding='utf-8') as file:
				self.queries[query_key] = file.read()

	def reset_queries(self):
		"""Reset the internal cache of queries, forcing queries to be reloaded upon next invocation"""
		self.queries=dict()
		self.prepared=dict()

	async def load_query(self, query_key):
		"""Load and return a single query by first looking it up in the internal cache and falling back on looking for it on disk"""
		# Cache hit
		if query_key in self.queries:
			return self.queries[query_key]
		# Cache miss, load from disk
		try:
			# Convert the query_key to a file path
			file_path = self.query_dir / pathlib.Path(query_key.replace('.', '/') + '.sql')
			logger.info(f"file_path {file_path}")
				
			# Read the SQL query from file
			async with aiofiles.open(file_path, mode='r', encoding='utf-8') as file:
				sql_query = await file.read()
			# with open(file_path, 'r', encoding='utf-8') as file:
			#	sql_query = file.read()
			self.queries[query_key] = sql_query
			return sql_query
		except FileNotFoundError:
			logger.info(f"SQL file {os.path.abspath(file_path)}")
			logger.error(f"SQL file not found for key {query_key} at path {file_path}")
		except Exception as e:
			logger.error(f"Error loading SQL file for key {query_key}: {e}")

	def _get_pool(self):
		self.pool = self.pool or psycopg_pool.AsyncConnectionPool(self.db_uri) # , min_size=min_conn, max_size=max_conn
		return self.pool

	async def direct_query(self, query, params=dict(), mode="none"):
		try:
			if not query:
				raise Exception("No query")
			# Make sure we actually have a pool
			pool = self._get_pool()
			if not pool:
				raise Exception("No pool")
			# TODO: Look into implementing pipeline mode
			# See: https://www.psycopg.org/psycopg3/docs/advanced/pipeline.html
			# Acquire a connection from the pool
			async with pool.connection() as connection:
				async with connection.transaction() as transaction:
					async with connection.cursor(row_factory=psycopg.rows.dict_row) as cursor:
						# See https://www.psycopg.org/psycopg3/docs/advanced/prepare.html
						logger.info(f"Preparing query with data: '{pprint.pformat(params)}'")
						await cursor.execute(query = query, params = params, prepare = True)
						if mode == "many":
							res = await cursor.fetchall()
							return res, None
						elif mode == "one":
							return await cursor.fetchone(), None
						elif mode == "exists":
							res = await cursor.fetchone(), None
							return res is not None
						elif mode == "none":
							return None, None
						else:
							raise Exception(f"Unknown mode {mode}. Should be one of [many, one, exists, none]")
		except Exception as e:
			logger.error("\n\n"+stripes())
			logger.error("")
			logger.error("")
			logger.error(f"###############################################")
			logger.error(f"#    Querying: \n\n{query}\n")
			if cursor and query and "cursor" in query:
				logger.error(f"#    Injected: '{cursor.query.decode()}'")
			logger.error(f"#         For")
			logger.error(f"#         + mode =   '{mode}'")
			logger.error(f"#         + params = '{pprint.pformat(params)}'")
			logger.error(f"# Failed with:")
			if isinstance(e,psycopg.Error):
				logger.error(f"# + e.type:     '{type(e).__name__}'")
				logger.error(f"# + e.detail:   '{e}'")
				logger.error(f"# + e.sqlstate: '{e.sqlstate}'")
				if getattr(e, "pgconn", None):
					logger.error(f"# + e.pgconn:   '{e.pgconn}'")
				if getattr(e, "pgresult", None):
					logger.error(f"# + e.pgresult: '{e.pgresult}'")
				if e.diag:
					logger.error(f"# + e.diag: '{e.diag}'")
					error_attrs = [
						"column_name", "constraint_name", "context", "datatype_name",
						"internal_position", "internal_query", "message_detail", "message_hint",
						"message_primary", "schema_name", "severity", "source_file",
						"source_function", "source_line", "sqlstate", "statement_position",
						"table_name", "severity_nonlocalized"
					]
					for attr in error_attrs:
						value = getattr(e.diag, attr, None)  # Use getattr to safely handle missing attributes
						if value is not None:  # Only log if the attribute is present and not None
							logger.error(f"#   + {attr}: {value}")
			else:
				# Unexpected errors
				logger.info(f"Unexpected error: {type(e).__name__}: {e}")
				
			logger.error("\n\n"+stripes())
			return None, str(e)


	async def _key_query_worker(self, query_key, params=dict(), mode="none"):
		query = await self.load_query(query_key)
		logger.info("_key_query_worker params='{pprint.pformat(params)}'")
		return await self.direct_query(query = query, params = params, mode = mode)

	async def query_many(self, query_key, params):
		logger.info("query_many params='{pprint.pformat(params)}'")
		return await self._key_query_worker(query_key = query_key, params = params, mode = "many")

	async def query_one(self, query_key, params):
		logger.info("query_one params='{pprint.pformat(params)}'")
		return await self._key_query_worker(query_key = query_key, params = params, mode = "one")

	async def query_none(self, query_key, params):
		logger.info("query_none params='{pprint.pformat(params)}'")
		return await self._key_query_worker(query_key = query_key, params = params, mode = "none")

	async def query_exists(self, query_key, params):
		logger.info("query_exists params='{pprint.pformat(params)}'")
		return await self._key_query_worker(query_key = query_key, params = params, mode = "exists")


	async def execute_old(self, query_key, **params):
		"""Execute a registered query by its key."""
		if query_key not in self.queries:
			raise ValueError(f"Query '{query_key}' not found.")
		
		async with self.pool.connection() as connection:
			async with connection.cursor() as cur:
				await cur.execute(self.queries[query_key], params)
				# Assuming a fetch is needed, adjust as necessary
				return await cur.fetchall()



import logging
import octomy.db
import octomy.web.context
import os
import pprint
import pytest
import datetime


logger = logging.getLogger(__name__)

# fmt: off
dummy_config= {
	"db-hostname":"hello.com"
	,"db-port":"1234"
	,"db-username":"arnold"
	,"db-password":"secret123"
	,"db-database":"mydb"
}
# fmt: on



@pytest.mark.asyncio
async def test_get_context():
	context = octomy.web.context.get_context()
	context = octomy.web.context.get_context(required_scopes = list())
	
	optional_key = "optional_key"
	optional_data = "optional_data"
	'''
	data:Data = Data()
	parts:Parts = Parts()
	search_engine:SearchEngine = SearchEngine()
	data:Data = None
	parts:Parts = None
	search_engine:SearchEngine = None
	'''
	outcome = None
	try:
		outcome = context.optional_key
	except:
		pass
	assert None is outcome, "optional_key should not be set at this point"
	outcome = None
	context.set_optional_attribute(optional_key, optional_data)
	try:
		outcome = context.optional_key
	except:
		pass
	assert optional_data is outcome, f"optional_key should be set to '{optional_data}' at this point"




#!/usr/bin/env python
import os
import posixpath
import hashlib
from flask import current_app
import logging
import octomy

logger = logging.getLogger(__name__)


# from https://github.com/pallets/flask/blob/024f0d384cf5bb65c76ac59f8ddce464b2dc2ca1/src/flask/helpers.py

# what separators does this operating system provide that are not a slash?
# this is used by the send_from_directory function to ensure that nobody is
# able to access files from outside the filesystem.
_os_alt_seps = list(sep for sep in [os.path.sep, os.path.altsep] if sep not in (None, "/"))


def safe_join(directory, *pathnames):
	"""Safely join `directory` and zero or more untrusted `pathnames`
	components.
	Example usage::
		@app.route('/wiki/<path:filename>')
		def wiki_page(filename):
			filename = safe_join(app.config['WIKI_FOLDER'], filename)
			with open(filename, 'rb') as fd:
				content = fd.read()  # Read and process the file content...
	:param directory: the trusted base directory.
	:param pathnames: the untrusted pathnames relative to that directory.
	:raises: :class:`~werkzeug.exceptions.NotFound` if one or more passed
			paths fall out of its boundaries.
	"""

	parts = [directory]

	for fn in pathnames:
		if fn != "":
			pre = fn
			fn = posixpath.normpath(fn)
			# logger.info(f"fn={fn}, pre={pre}")
		if any(sep in fn for sep in _os_alt_seps) or os.path.isabs(fn) or fn == ".." or fn.startswith("../"):
			# logger.info(f"pathname was invalid: {fn}")
			return None

		parts.append(fn)

	return posixpath.join(*parts)


expiry_cache = {}


def get_expiry(path_raw):
	direc = current_app.config["static-files-root"]
	static_prefix = "/static/"
	path = path_raw[len(static_prefix) :] if path_raw.startswith(static_prefix) else path_raw
	filename_temp = os.fspath(path)
	directory = os.fspath(direc)
	filename = safe_join(directory, filename_temp)
	# logger.info(f"path={path}, direc={direc}, filename_temp={filename_temp}, directory={directory}, filename={filename}")
	if not filename:
		# logger.info("No filename returned from safejoin")
		return path_raw
	if not os.path.isabs(filename):
		# logger.info(f"{filename} was not abolute")
		filename = os.path.join(current_app.root_path, filename)
	try:
		if os.path.isfile(filename):
			if not filename in expiry_cache:
				m = hashlib.sha256(octomy.utils.read_file(filename).encode("utf-8"))
				expiry_cache[filename] = m.hexdigest()[0:6]
			return f"{path_raw}?expiry_hash={expiry_cache[filename]}"
		else:
			# logger.info(f"filename={filename} was not afile")
			pass
	except Exception as e:
		# logger.info(f"Exception:{e}", exc_info=True)
		pass
	return path_raw



from threading import Timer
import logging

logger = logging.getLogger(__name__)


class Watchdog(BaseException):
	def __init__(self, timeout, userHandler=None):  # timeout in seconds
		self.timeout = timeout
		# logger.info(f"WATCHDOG SET WITH TIME {self.timeout}")
		self.handler = userHandler if userHandler is not None else self.defaultHandler
		self.timer = Timer(self.timeout, self.handler)
		self.timer.start()

	def reset(self):
		self.timer.cancel()
		self.timer = Timer(self.timeout, self.handler)
		self.timer.start()

	def stop(self):
		# logger.info(f"WATCHDOG STOPPED")
		self.timer.cancel()

	def defaultHandler(self):
		logger.info("WATCHDOG TIMED OUT")
		raise self



import json
import os
import sqlite3
import pprint
import psycopg
import traceback
import sys
import logging
from urllib.parse import urlparse, urlunparse, ParseResult
import pytz
import fk.utils
from .utils import config_to_url, get_config_hash
import unicodedata

import pprint

logger = logging.getLogger(__name__)


class DatabaseConnectionAsync:

    def __init__(self, config, event_source=None):
        self.config = config
        self.event_source=event_source
        self.db_hostname = self.config.get("db-hostname", None)
        self.db_port = self.config.get("db-port", None)
        self.db_username = self.config.get("db-username", None)
        self.db_password = self.config.get("db-password", None)
        self.db_database = self.config.get("db-database", None)
        self.db = None
        self.connstr = config_to_url(config, do_redact=False)
        connstr_safe= config_to_url(config, do_redact=True)
        logger.info(f"DATABASE(ASYNC): {connstr_safe}")
        self.pool=ConnectionPool(conninfo=self.connstr, open=not self.event_source, min_size=1, max_size = 100)
        if self.event_source:
            self.add_event_handler("startup", pool.open)
            self.add_event_handler("shutdown", pool.close)

        self.connection_error = None
        self._prepare_db()

    def __del____(self):
        self._unprepare_db()

    def is_ok(self):
        return self.db_hostname and self.db_port and self.db_username and self.db_password and self.db_database

    async def _unprepare_db(self):
        if self.db:
            try:
                # Make sure data is stored
                self.db.commit()
                self.db.close()
            except Exception as e1:
                pass
            try:
                del self.db
            except Exception as e2:
                pass
            self.db = None
            logger.info("PostgreSQL connection is closed")

    # Internal helper to connect to database
    async def _prepare_db(self):
        with await psycopg.AsyncConnection.connect(conninfo=self.connstr) as aconn:
           async with aconn.cursor() as acur:
               await acur.execute("SELECT now();")
               await acur.fetchone()
               async for record in acur:
                   logger.info(f"NOW ASYNC: {record}")

"""
            self.db = psycopg2.connect(host=self.db_hostname, port=self.db_port, user=self.db_username, password=self.db_password, database=self.db_database)

            # Create a cursor to let us work with the database
            with self.db:
                with self.db.cursor() as c:
                    # logger.info( self.db.get_dsn_parameters(),"\n")

                    c.execute("SELECT version();")
                    record = c.fetchone()
                    logger.info(f"Connected to: {record[0]}\n")
                    return True

        except (Exception, psycopg2.Error) as error:
            logger.warning("###   ###   ###   ###   ###   ###   ###   #")
            logger.warning("##   ###   ###   ###   ###   ###   ###   ##")
            logger.warning("")
            logger.warning(f"Error while connecting to PostgreSQL {error}")
            self.connection_error = error
            self._unprepare_db()
        return False

    async def reconnect(self):
        logger.info("###   ###   ###   ###   ###   ###   ###   #")
        logger.info("##   ###   ###   ###   ###   ###   ###   ##")
        logger.info("")
        logger.info(f"Reconnecting...")
        await self._unprepare_db()
        return await self._prepare_db()
"""



from dotenv import dotenv_values
from octomy.cad import openscad
import datetime
import logging
import os
import pprint
import pytest
import re
import shutil

logger = logging.getLogger(__name__)




here = os.path.dirname(__file__)
data_dir = os.path.join(here, "data")
env = dict()
env_re = re.compile(r'export (?P<key>.*)=(?P<value>.*)')


def parse_env(here):
	env_file = os.path.join(here, ".env")
	return dotenv_values(env_file)

def parse_env_old(here):
	env_file = os.path.join(here, ".env")
	out = dict()
	if os.path.isfile(env_file):
		with open(env_file, "r") as f:
			data = f.read()
			#logger.info(f"data: {data}")
			for m in env_re.finditer(data):
				gd = m.groupdict()
				#logger.info(f"gd: {gd}")
				out[gd.get('key')] = gd.get('value')
				#logger.info(f"line: {gd.get('key')} = {gd.get('value')}")
	else:
		logger.warning(f"Could not find env file: {env_file}")
	return out


def get_env(key, fb=None):
	global env
	if env == {}:
		env = parse_env(here)
	ret = env.get(key) or os.environ.get(key) or fb
	logger.info(f"GOT {key}={ret}")
	return ret


def remove_dir(dir):
	logger.info(f"Removing dir '{dir}'")
	if os.path.isdir(dir):
		shutil.rmtree(dir, ignore_errors=True)
	assert(not os.path.isdir(dir))

"""
@pytest.fixture(scope='function')
def git_repo(request):
	logger.info(f"here:{here}, project_folder:{git_folder}")
	git_data = {
		  "path": "github.com/arribatec-cloud/small-test-repo.git"
		, "username": get_env("TEST_GIT_USERNAME", None)
		, "password": get_env("TEST_GIT_PASSWORD", None)
		, "do_debug": True
	}
	logger.info("creating client with:")
	logger.info(pprint.pformat(git_data))
	remove_dir(git_folder)
	repo = git.Client(**git_data)
	logger.info("creating client DONE")
	assert(repo)
	repo.state()
	def teardown():
		logger.info("TEARDONE")
		repo = None
		remove_dir(git_folder)
	request.addfinalizer(teardown)
	return repo
"""


expected_schema = [{'default': {'type': 'number', 'value': 1},
'description': 'DIN-780 standard tooth height',
'name': 'modul',
'schema': {'raw': '[0.05, 0.06, 0.08, 0.10, 0.12, '
				  '0.16,0.20,0.25,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1,1.25,1.5,2,2.5,3,4,5,6,8,10,12,16,20,25,32,40,50,60]',
		   'values': [0.05,
					  0.06,
					  0.08,
					  0,
					  0.12,
					  0.16,
					  0,
					  0.25,
					  0.3,
					  0.4,
					  0.5,
					  0.6,
					  0.7,
					  0.8,
					  0.9,
					  1,
					  1.25,
					  1.5,
					  2,
					  2.5,
					  3,
					  4,
					  5,
					  6,
					  8,
					  10,
					  12,
					  16,
					  20,
					  25,
					  32,
					  40,
					  50,
					  60]},
'section': 'Spur Gear'},
{'default': {'type': 'number', 'value': 20},
'description': 'Number of teeth',
'name': 'tooth_count',
'schema': {'max': 30, 'min': 2, 'raw': '[2:30]', 'step': 0},
'section': 'Spur Gear'},
{'default': {'type': 'number', 'value': 40},
'description': 'Length',
'name': 'length',
'schema': {'max': 100, 'min': 1, 'raw': '[1:100]', 'step': 0},
'section': 'Spur Gear'},
{'default': {'type': 'number', 'value': 4},
'description': 'Height',
'name': 'height',
'schema': {'max': 100, 'min': 1, 'raw': '[1:100]', 'step': 0},
'section': 'Spur Gear'},
{'default': {'type': 'number', 'value': 8},
'description': 'Width',
'name': 'width',
'schema': {'max': 100, 'min': 1, 'raw': '[1:100]', 'step': 0},
'section': 'Spur Gear'},
{'default': {'type': 'number', 'value': 2},
'description': 'Bore diameter',
'name': 'bore',
'schema': {'label': 'default',
		   'labels': {2: 'default', '0': 80},
		   'raw': '[0:80]'},
'section': 'Spur Gear'},
{'default': {'type': 'bool', 'value': True},
'description': 'Remove uneccessary material from gears',
'name': 'optimize',
'schema': {'debug': 'NO SCHEMA FOR None with type bool'},
'section': 'Spur Gear'},
{'default': {'type': 'number', 'value': 30},
'description': 'Attack angle',
'name': 'attack_angle',
'schema': {'label': 'default',
		   'labels': {30: 'default', '0': 30},
		   'raw': '[0:30]'},
'section': 'Teeth'},
{'default': {'type': 'number', 'value': 20},
'description': 'Helix angle (for helix and herringbone teeth)',
'name': 'helix_angle',
'schema': {'label': 'default',
		   'labels': {20: 'default', '0': 30},
		   'raw': '[0:30]'},
'section': 'Teeth'},
{'default': {'type': 'bool', 'value': False},
'description': 'Use herring bone teeth',
'name': 'herringbone',
'schema': {'debug': 'NO SCHEMA FOR None with type bool'},
'section': 'Teeth'},
{'default': {'type': 'bool', 'value': True},
'description': 'Include pinion gear',
'name': 'pair',
'schema': {'debug': 'NO SCHEMA FOR None with type bool'},
'section': 'Pinion Gear'},
{'default': {'type': 'bool', 'value': True},
'description': 'Show parts as assembled',
'name': 'assembled',
'schema': {'debug': 'NO SCHEMA FOR None with type bool'},
'section': 'Pinion Gear'},
{'default': {'type': 'number', 'value': 2},
'description': 'Bore diameter',
'name': 'bore_secondary',
'schema': {'label': 'default',
		   'labels': {2: 'default', '0': 80},
		   'raw': '[0:80]'},
'section': 'Pinion Gear'},
{'default': {'type': 'number', 'value': 10},
'description': 'Number of teeth',
'name': 'tooth_count_secondary',
'schema': {'max': 30, 'min': 2, 'raw': '[2:30]', 'step': 0},
'section': 'Pinion Gear'}]


def test_openscad_parse():
	logger.info("openscad parse unit test")
	local_name_openscad = os.path.join(data_dir, "customizer.scad")
	with open(local_name_openscad, "r") as f:
		source = f.read()
		schema = openscad.parse_customize(source)
		logger.info("SCHEMA:")
		logger.info(pprint.pformat(schema))
		#assert(expected_schema == schema)
		logger.info("SOURCE:")
		logger.info(source)

	
def test_openscad_render():
	logger.info("openscad render unit test")
	do_debug = True
	set_name = "set"
	local_name_openscad = os.path.join(data_dir, "Spur.scad")
	local_name_stl = f"{local_name_openscad}.stl"
	local_name_png = f"{local_name_openscad}.png"
	local_name_json = f"{local_name_openscad}.json"
	#json_parameters = json.dumps(openscad_parameters, indent=3, sort_keys=True)
	logger.info(f"do_debug:{do_debug}")
	logger.info(f"set_name:{set_name}")
	logger.info(f"local_name_openscad:{local_name_openscad}")
	logger.info(f"local_name_stl:{local_name_stl}")
	logger.info(f"local_name_json:{local_name_json}")
	stl_scad = openscad.OpenScadRunner(
		scriptfile = local_name_openscad
		, outfiles = [local_name_stl, local_name_png]
		, customizer_file = local_name_json
		, customizer_sets = [set_name]
		, auto_center = True
		, view_all = True
		, verbose = do_debug or True)
	ok, err = stl_scad.run()
	logger.info(f"success:{stl_scad.success}")
	logger.info(f"return_code:{stl_scad.return_code}")
	logger.info(f"OK:{ok}")
	logger.info(f"err:{err}")
	return True




test_cases = [
	{
		"markup": """
/* [Drop down box] */
// combo box for number
Numbers=2; // [0, 1, 2, 3]
	""",
		"expected": None,
	},
	{
		"markup": """
/* [Drop down box] */
// combo box for string
Strings="foo"; // [foo, bar, baz]
""",
		"expected": [{'default': {'type': 'string', 'value': 'foo'},
		'description': 'combo box for string',
		'name': 'Strings',
		'schema': {'raw': '[foo, bar, baz]', 'values': ['foo', 'bar', 'baz']},
		'section': 'Drop down box'
		}],
	},
	{
		"markup": """
/* [Drop down box] */
// labeled combo box for numbers
Labeled_values=10; // [10:L, 20:M, 30:XL]
""",
		"expected": [{'default': {'type': 'number', 'value': '10'},
		'description': 'labeled combo box for numbers',
		'name': 'Labeled_values',
		'schema': {'labels': {10: 'L', 20: 'M', 30: 'XL'},
				   'raw': '[10:L, 20:M, 30:XL]'},
		'section': 'Drop down box',
		}],
	},
	{
		"markup": """
/* [Drop down box] */
// labeled combo box for string
Labeled_value="S"; // [S:Small, M:Medium, L:Large]
""",
		"expected": 
			  [{'default': {'type': 'string', 'value': 'S'}
			, 'description': 'labeled combo box for string'
			, 'name': 'Labeled_value'
			, 'schema': {'labels': {'L': 'Large', 'M': 'Medium', 'S': 'Small'},
						 'raw': '[S:Small, M:Medium, L:Large]'},
			  'section': 'Drop down box'}]
	},
	{
		"markup": """
/*[ Slider ]*/
// slider widget for number
slider =34; // [10:100]
""",
		"expected": None,
	},
	{
		"markup": """
/*[ Slider ]*/
// step slider for number
stepSlider=2; //[0:5:100]
""",
		"expected": None,
	},
	{
		"markup": """
/* [Checkbox] */
// description
Variable = true;
""",
		"expected": None,
	},
	{
		"markup": """
/* [Spinbox] */
// spinbox with step size 1
Spinbox1 = 5;
""",
		"expected": None,
	},
	{
		"markup": """
/* [Spinbox with unit] */
// spinbox with step size 1 and unit
Spinbox2 = 5; // (mm)
		""",
		"expected": None,
	},
	{
		"markup": """
/* [Textbox] */
// Text box for vector with more than 4 elements
Vector6=[12,34,44,43,23,23];
""",
		"expected": None,
	},
	{
		"markup": """
/* [Textbox] */
// Text box for string
String="hello";
""",
		"expected": None,
	},
	{
		"markup": """
/* [Vector] */
// Text box for vector with 1 element
Vector1=[12]; //[0:2:50]
""",
		"expected": None,
	},
	{
		"markup": """
/* [Vector] */
// Text box for vector with 2 elements
Vector2=[12,34]; //[0:2:50]
""",
		"expected": None,
	},
	{
		"markup": """
/* [Vector] */
// Text box for vector with 3 elements
Vector3=[12,34,46]; //[0:2:50]
""",
		"expected": None,
	},
	{
		"markup": """
/* [Vector] */
// Text box for vector with 4 elements
Vector4=[12,34,46,24]; //[0:2:50]
""",
		"expected": None,
	},
]


def lol_test_openscad_parse_cases():
	for test_case in test_cases:
		source = test_case.get("markup")
		expected_schema = test_case.get("expected")
		schema = openscad.parse_customize(source, do_debug = False)
		if expected_schema:
			logger.info("-----------------------------------------------------\n"+source+"\n")
			logger.info("SCHEMA:\n\n"+pprint.pformat(schema)+"\n")
			logger.info("EXPECTED:\n\n"+pprint.pformat(expected_schema)+"\n")
			assert(expected_schema == schema)


regexes = {
"decimal_number_re":{
		"regex": re.compile(openscad.decimal_number_inc)
		, "tests":{
			  ".0": {}
			  , "": None
			  , "X": None
			, "0.0": {}
			, "-0.0": {}
		}
	}
	, "quoted_string_re":{
		"regex": re.compile(openscad.quoted_string_inc)
		, "tests":{
			  "'bob'": {'value': 'bob'}
			, "\"lol\"": {'value': 'lol'}
			, "''": {'value': ''}
			, "\"\"": {'value': ''}
			, "'\"": None
			, "'123345456567679780qwertyuiopåæasdfghjkløæzxcvbnm,.!\"#¤#¤%&%&/&/()/()=()=?'":  {'value': '123345456567679780qwertyuiopåæasdfghjkløæzxcvbnm,.!"#¤#¤%&%&/&/()/()=()=?'}
		}
	}
	, "unit_re":{
		"regex": re.compile(openscad.unit_inc)
		, "tests":{
			  "(mm)": {'unit': 'mm'}
			, "()": None
		}
	}
	, "h1_re":{
		"regex": openscad.h1_re
		, "tests":{
			  "/* [Section name] */": {'section': 'Section name'}
			, "/* [Section name] */": {'section': 'Section name'}
			, "/* [ Section name] */": {'section': 'Section name'}
			, "/* [Section name ] */": {'section': 'Section name'}
			, "/* [ Section name ] */": {'section': 'Section name'}
			, "/* [Section name]  */": {'section': 'Section name'}
			, "/*  [Section name] */": {'section': 'Section name'}
			, "/*[Section name] */": {'section': 'Section name'}
			, "/* [Section name]*/": {'section': 'Section name'}
			, "/*[Section name]*/": {'section': 'Section name'}
			, "/*[e]*/": {'section': 'e'}
			, "/* [ e ] */": {'section': 'e'}
			, "/*[]*/": None
			, "/* [ ] */": None
			, "/* [ e ] * /": None
			, "/* [ ] * /": None
			, "": None
		}
	}
	, "h2_re":{
		"regex": openscad.h2_re
		, "tests":{
			  "// Variable description": {'description': 'Variable description', 'unit': None}
			, " // Variable description": {'description': 'Variable description', 'unit': None}
			, "// Variable description": {'description': 'Variable description', 'unit': None}
			, "//  Variable description": {'description': 'Variable description', 'unit': None}
			, "/ / Variable description": None
			, "/ /Variable description": None
			, "// Variable description ": {'description': 'Variable description', 'unit': None}
			, "// Variabledescription": {'description': 'Variabledescription', 'unit': None}
			, " // Variable description ": {'description': 'Variable description', 'unit': None}
			, " // Variabledescription": {'description': 'Variabledescription', 'unit': None}
			, "/ / Variable description ": None
			, "/ / Variabledescription": None
			, " / / Variable description ": None
			, " / / Variabledescription": None
			, "//": None
		}
	}
	, "var_re":{
		"regex": openscad.var_re
		, "tests":{
			  ".0": None
			, "Variable = definition": None
			, "Variable = definition;": {'variable': 'Variable', 'value': 'definition', 'comment': None}
			, "lolbob = badabing; // comment": {'comment': 'comment', 'value': 'badabing', 'variable': 'lolbob'}
			, "  lolbob   =   badabing  ; //    comment": {'comment': 'comment', 'value': 'badabing', 'variable': 'lolbob'}
			, "lolbob=badabing;//comment": {'comment': 'comment', 'value': 'badabing', 'variable': 'lolbob'}
			, "  lolbob   =   badabing  ; / /    comment": None
			, "lolbob = badabing;": {'comment': None, 'value': 'badabing', 'variable': 'lolbob'}
			, "  lolbob = badabing;": {'comment': None, 'value': 'badabing', 'variable': 'lolbob'}
			, "  lolbob   = badabing  ;   ": {'comment': None, 'value': 'badabing', 'variable': 'lolbob'}
			, "lolbob=badabing;": {'comment': None, 'value': 'badabing', 'variable': 'lolbob'}
			, "lolbob = badabing //": None
			, "lolbob = badabing": None
			, "lolbob = badabing  ": None
			, "  lolbob = badabing": None
			, "lolbob=badabing": None
			, "lolbob = badabing;": {'comment': None, 'value': 'badabing', 'variable': 'lolbob'}
			, "X": None
			, "0.0": None
			, "-0.0": None
		}
	}
	, "bool_val_re":{
		"regex": openscad.bool_val_re
		, "tests":{
			  ".0": None
			, "": None
			, " false ": {'value': 'false'}
			, " false": {'value': 'false'}
			, " true ": {'value': 'true'}
			, " true": {'value': 'true'}
			, "false ": {'value': 'false'}
			, "false": {'value': 'false'}
			, "true ": {'value': 'true'}
			, "true": {'value': 'true'}
			, "X": None
			, "0.0": None
			, "-0.0": None
		}
	}
	, "string_val_re":{
		"regex": openscad.string_val_re
		, "tests":{
			  ".0": None
			, "": None
			, "X": None
			, "0.0": None
			, "-0.0": None
			, "''": None
			, "\"\"": None
			, "X": None
			, "0.0": None
			, "-0.0": None
			, "'bob'": {'value': 'bob'}
			, "\"lol\"": {'value': 'lol'}
			, "''": {'value': ''}
			, "\"\"": {'value': ''}
			, "'\"": None
			, "'123345456567679780qwertyuiopåæasdfghjkløæzxcvbnm,.!\"#¤#¤%&%&/&/()/()=()=?'":  {'value': '123345456567679780qwertyuiopåæasdfghjkløæzxcvbnm,.!"#¤#¤%&%&/&/()/()=()=?'}
			, " 'bob'": {'value': 'bob'}
			, " \"lol\"": {'value': 'lol'}
			, " ''": {'value': ''}
			, " \"\"": {'value': ''}
			, " '\"": None
			, " '123345456567679780qwertyuiopåæasdfghjkløæzxcvbnm,.!\"#¤#¤%&%&/&/()/()=()=?'":  {'value': '123345456567679780qwertyuiopåæasdfghjkløæzxcvbnm,.!"#¤#¤%&%&/&/()/()=()=?'}
			, "'bob' ": {'value': 'bob'}
			, "\"lol\" ": {'value': 'lol'}
			, "'' ": {'value': ''}
			, "\"\" ": {'value': ''}
			, "'\" ": None
			, "'123345456567679780qwertyuiopåæasdfghjkløæzxcvbnm,.!\"#¤#¤%&%&/&/()/()=()=?' ":  {'value': '123345456567679780qwertyuiopåæasdfghjkløæzxcvbnm,.!"#¤#¤%&%&/&/()/()=()=?'}
			, " 'bob' ": {'value': 'bob'}
			, " \"lol\" ": {'value': 'lol'}
			, " '' ": {'value': ''}
			, " \"\" ": {'value': ''}
			, " '\" ": None
			, " '123345456567679780qwertyuiopåæasdfghjkløæzxcvbnm,.!\"#¤#¤%&%&/&/()/()=()=?' ":  {'value': '123345456567679780qwertyuiopåæasdfghjkløæzxcvbnm,.!"#¤#¤%&%&/&/()/()=()=?'}
		}
	}
	, "number_val_re":{
		"regex": openscad.number_val_re
		, "tests":{
			  ".0": {'value': '.0'}
			, "": None
			, "X": None
			, "0.0": {'value': '0.0'}
			, "-0.0": {'value': '-0.0'}
		}
	}
	, "labeled_bool_schema_re":{
		"regex": openscad.labeled_bool_schema_re
		, "tests":{
			  ".0": None
			, "-0.0": None
			, " [ true:label1, false:label2 ] ": {'values': 'true:label1, false:label2'}
			, "word": None
			, "number": None
			, "true": None
			, "false": None
			, "[]": None
			, "[ ]": None
			, "[word]": None
			, "[false]": None
			, "[true]": None
			, "[false:]": None
			, "[true:]": None
			, "[false :]": None
			, "[true :]": None
			, "[ false]": None
			, "[ true]": None
			, "[ false:]": None
			, "[ true:]": None
			, "[ false :]": None
			, "[ true :]": None
			, "[false ]": None
			, "[true ]": None
			, "[false: ]": None
			, "[true: ]": None
			, "[false : ]": None
			, "[true : ]": None
			, "[ false ]": None
			, "[ true ]": None
			, "[ false: ]": None
			, "[ true: ]": None
			, "[ false : ]": None
			, "[ true : ]": None
			, "[ true:label1, false:label2, false:label2, false:label2, false:label2]": {'values': 'true:label1, false:label2, false:label2, false:label2, false:label2'}
			, "[false:label2]": {'values': 'false:label2'}
			, "[true:label2]": {'values': 'true:label2'}
			, "[ false:label2]": {'values': 'false:label2'}
			, "[ true:label2]": {'values': 'true:label2'}
			, "[false:label2 ]": {'values': 'false:label2'}
			, "[true:label2 ]": {'values': 'true:label2'}
			, "[ false:label2 ]": {'values': 'false:label2'}
			, "[ true:label2 ]": {'values': 'true:label2'}
			, "[ false:label2,false:label2]": {'values': 'false:label2,false:label2'}
			, "[ true:label2,false:label2]": {'values': 'true:label2,false:label2'}
			, "[false:label2,false:label2 ]": {'values': 'false:label2,false:label2'}
			, "[true:label2,false:label2 ]": {'values': 'true:label2,false:label2'}
			, "[ false:label2,false:label2 ]": {'values': 'false:label2,false:label2'}
			, "[ true:label2,false:label2 ]": {'values': 'true:label2,false:label2'}
			, "[ false:label2, false:label2]": {'values': 'false:label2, false:label2'}
			, "[ true:label2, false:label2]": {'values': 'true:label2, false:label2'}
			, "[false:label2, false:label2 ]": {'values': 'false:label2, false:label2'}
			, "[true:label2, false:label2 ]": {'values': 'true:label2, false:label2'}
			, "[ false:label2, false:label2 ]": {'values': 'false:label2, false:label2'}
			, "[ true:label2, false:label2 ]": {'values': 'true:label2, false:label2'}
			, "[ false:label2 ,false:label2]": {'values': 'false:label2 ,false:label2'}
			, "[ true:label2 ,false:label2]": {'values': 'true:label2 ,false:label2'}
			, "[false:label2 ,false:label2 ]": {'values': 'false:label2 ,false:label2'}
			, "[true:label2 ,false:label2 ]": {'values': 'true:label2 ,false:label2'}
			, "[ false:label2 ,false:label2 ]": {'values': 'false:label2 ,false:label2'}
			, "[ true:label2 ,false:label2 ]": {'values': 'true:label2 ,false:label2'}
			, "[ false:label2 , false:label2]": {'values': 'false:label2 , false:label2'}
			, "[ true:label2 , false:label2]": {'values': 'true:label2 , false:label2'}
			, "[false:label2 , false:label2 ]": {'values': 'false:label2 , false:label2'}
			, "[true:label2 , false:label2 ]": {'values': 'true:label2 , false:label2'}
			, "[ false:label2 , false:label2 ]": {'values': 'false:label2 , false:label2'}
			, "[ true:label2 , false:label2 ]": {'values': 'true:label2 , false:label2'}
		}
	} 
	, "values_string_schema_re":{
		"regex": openscad.values_string_schema_re
		, "tests":{
			  "": None
			, " ": None
			, "[]": None
			, " [ ] ": None
			, " [] ": None
			, "[] ": None
			, " []": None
			, "[]": None
			, "[ value1, value2, value3 ]": {'values': 'value1, value2, value3'}
			, "[value1,value2,value3]": {'values': 'value1,value2,value3'}
			, "[value1, value2, value3 ]": {'values': 'value1, value2, value3'}
			, "[ value1 , value2 , value3 ]": {'values': 'value1 , value2 , value3'}
			, "[value1:, value2, value3 ]": None
			, "[ value1: , value2 , value3 ]": None
		}
	}
	, "labeled_string_schema_re":{
		"regex": openscad.labeled_string_schema_re
		, "tests":{
			  "": None
			, " ": None
			, "[]": None
			, " [ ] ": None
			, " [] ": None
			, "[] ": None
			, " []": None
			, "[]": None
			, "[ value1, value2, value3 ]": None
			, "[value1,value2,value3]": None
			, "[value1, value2, value3 ]": None
			, "[ value1 , value2 , value3 ]": None
			, "[value1:, value2, value3 ]": None
			, "[ value1: , value2 , value3 ]": None
			, "[value1 :label1,value2: label2,value3:label3]": {'values': 'value1 :label1,value2: label2,value3:label3'}
			, "[value1 :label1, value2: label2, value3:label3]": {'values': 'value1 :label1, value2: label2, value3:label3'}
			, "[value1 :label1 ,value2: label2 ,value3:label3]": {'values': 'value1 :label1 ,value2: label2 ,value3:label3'}
			, "[value1 :label1 , value2: label2 , value3:label3]": {'values': 'value1 :label1 , value2: label2 , value3:label3'}
			, "[ value1 :label1,value2: label2,value3:label3]": {'values': 'value1 :label1,value2: label2,value3:label3'}
			, "[ value1 :label1 ,value2: label2 ,value3:label3]": {'values': 'value1 :label1 ,value2: label2 ,value3:label3'}
			, "[ value1 :label1, value2: label2, value3:label3]": {'values': 'value1 :label1, value2: label2, value3:label3'}
			, "[ value1 :label1 , value2: label2 , value3:label3]": {'values': 'value1 :label1 , value2: label2 , value3:label3'}
			, "[value1 :label1,value2: label2,value3:label3 ]": {'values': 'value1 :label1,value2: label2,value3:label3'}
			, "[value1 :label1, value2: label2, value3:label3 ]": {'values': 'value1 :label1, value2: label2, value3:label3'}
			, "[value1 :label1 ,value2: label2 ,value3:label3 ]": {'values': 'value1 :label1 ,value2: label2 ,value3:label3'}
			, "[value1 :label1 , value2: label2 , value3:label3 ]": {'values': 'value1 :label1 , value2: label2 , value3:label3'}
			, "[ value1 :label1,value2: label2,value3:label3 ]": {'values': 'value1 :label1,value2: label2,value3:label3'}
			, "[ value1 :label1, value2: label2, value3:label3 ]": {'values': 'value1 :label1, value2: label2, value3:label3'}
			, "[ value1 :label1 ,value2: label2 ,value3:label3 ]": {'values': 'value1 :label1 ,value2: label2 ,value3:label3'}
			, "[ value1 :label1 , value2: label2 , value3:label3 ]": {'values': 'value1 :label1 , value2: label2 , value3:label3'}
		}
	}
	, "range1_schema_re":{
		"regex": openscad.range1_schema_re
		, "tests":{
			  "": None
			, " ": None
			, "[]": None
			, " [ ] ": None
			, " [] ": None
			, "[] ": None
			, " []": None
			, "[]": None
			, "[0]": {'max': '0', 'unit': None}
			, "[-0]": {'max': '-0', 'unit': None}
			, "[1]": {'max': '1', 'unit': None}
			, "[-1]": {'max': '-1', 'unit': None}
			, "[.0]": {'max': '.0', 'unit': None}
			, "[-.0]": {'max': '-.0', 'unit': None}
			, "[-0.0]": {'max': '-0.0', 'unit': None}
			, "[-1.0]": {'max': '-1.0', 'unit': None}
			, "[12345.67890]": {'max': '12345.67890', 'unit': None}
			, "[-12345.67890]": {'max': '-12345.67890', 'unit': None}
			, "[0 ]": {'max': '0', 'unit': None}
			, "[-0 ]": {'max': '-0', 'unit': None}
			, "[1 ]": {'max': '1', 'unit': None}
			, "[-1 ]": {'max': '-1', 'unit': None}
			, "[.0 ]": {'max': '.0', 'unit': None}
			, "[-.0 ]": {'max': '-.0', 'unit': None}
			, "[-0.0 ]": {'max': '-0.0', 'unit': None}
			, "[-1.0 ]": {'max': '-1.0', 'unit': None}
			, "[12345.67890 ]": {'max': '12345.67890', 'unit': None}
			, "[-12345.67890 ]": {'max': '-12345.67890', 'unit': None}
			, "[ 0]": {'max': '0', 'unit': None}
			, "[ -0]": {'max': '-0', 'unit': None}
			, "[ 1]": {'max': '1', 'unit': None}
			, "[ -1]": {'max': '-1', 'unit': None}
			, "[ .0]": {'max': '.0', 'unit': None}
			, "[ -.0]": {'max': '-.0', 'unit': None}
			, "[ -0.0]": {'max': '-0.0', 'unit': None}
			, "[ -1.0]": {'max': '-1.0', 'unit': None}
			, "[ 12345.67890]": {'max': '12345.67890', 'unit': None}
			, "[ -12345.67890]": {'max': '-12345.67890', 'unit': None}
			, "[ 0 ]": {'max': '0', 'unit': None}
			, "[ -0 ]": {'max': '-0', 'unit': None}
			, "[ 1 ]": {'max': '1', 'unit': None}
			, "[ -1 ]": {'max': '-1', 'unit': None}
			, "[ .0 ]": {'max': '.0', 'unit': None}
			, "[ -.0 ]": {'max': '-.0', 'unit': None}
			, "[ -0.0 ]": {'max': '-0.0', 'unit': None}
			, "[ -1.0 ]": {'max': '-1.0', 'unit': None}
			, "[ 12345.67890 ]": {'max': '12345.67890', 'unit': None}
			, "[ -12345.67890 ]": {'max': '-12345.67890', 'unit': None}
			, "[ 1 ]()": {'max': '1', 'unit': None}
			, "[ 1 ] ()": {'max': '1', 'unit': None}
			, "[ 1 ]( )": {'max': '1', 'unit': None}
			, "[ 1 ]() ": {'max': '1', 'unit': None}
			, "[ 1 ] ( ) ": {'max': '1', 'unit': None}
			, "[ 1 ](mm)": {'max': '1', 'unit': 'mm'}
			, "[ 1 ]( mm)": {'max': '1', 'unit': 'mm'}
			, "[ 1 ]( mm )": {'max': '1', 'unit': 'mm'}
			, "[ 1 ](mm )": {'max': '1', 'unit': 'mm'}
			, "[ 1 ] ( mm )": {'max': '1', 'unit': 'mm'}
			, "[ 1 ] (mm )": {'max': '1', 'unit': 'mm'}
			, "[ 1 ] ( mm)": {'max': '1', 'unit': 'mm'}
			, "[ 1 ] (mm)": {'max': '1', 'unit': 'mm'}
			, "[ 1 ](mm) ": {'max': '1', 'unit': 'mm'}
			, "[ 1 ]( mm) ": {'max': '1', 'unit': 'mm'}
			, "[ 1 ]( mm ) ": {'max': '1', 'unit': 'mm'}
			, "[ 1 ](mm ) ": {'max': '1', 'unit': 'mm'}
			, "[ 1 ] ( mm ) ": {'max': '1', 'unit': 'mm'}
			, "[ 1 ] (mm ) ": {'max': '1', 'unit': 'mm'}
			, "[ 1 ] ( mm) ": {'max': '1', 'unit': 'mm'}
			, "[ 1 ] (mm) ": {'max': '1', 'unit': 'mm'}

		}
	}
	, "range2_schema_re":{
		"regex": openscad.range2_schema_re
		, "tests":{
			  "": None
			, " ": None
			, "[]": None
			, " [ ] ": None
			, " [] ": None
			, "[] ": None
			, " []": None
			, "[]": None
			, "[0]": None
			, "[-0]": None
			, "[1]": None
			, "[-1]": None
			, "[.0]": None
			, "[-.0]": None
			, "[-0.0]": None
			, "[-1.0]": None
			, "[12345.67890]": None
			, "[-12345.67890]": None
			, "[ 1 ](mm)": None
			, "[ 1 ]( mm)": None
			, "[ 1 ]( mm )": None
			, "[ 1 ](mm )": None
			, "[ 1 ] ( mm )": None
			, "[ 1 ] (mm )": None
			, "[ 1 ] ( mm)": None
			, "[ 1 ] (mm)": None
			, "[ 1 ](mm) ": None
			, "[ 1 ]( mm) ": None
			, "[ 1 ]( mm ) ": None
			, "[ 1 ](mm ) ": None
			, "[ 1 ] ( mm ) ": None
			, "[ 1 ] (mm ) ": None
			, "[ 1 ] ( mm) ": None
			, "[ 1 ] (mm) ": None
			, "[0:0]": {'min': '0', 'max': '0', 'unit': None}
			, "[-0:0]": {'min': '-0', 'max': '0', 'unit': None}
			, "[1:0]": {'min': '1', 'max': '0', 'unit': None}
			, "[-1:0]": {'min': '-1', 'max': '0', 'unit': None}
			, "[.0:0]": {'min': '.0', 'max': '0', 'unit': None}
			, "[-.0:0]": {'min': '-.0', 'max': '0', 'unit': None}
			, "[-0.0:0]": {'min': '-0.0', 'max': '0', 'unit': None}
			, "[-1.0:0]": {'min': '-1.0', 'max': '0', 'unit': None}
			, "[12345.67890:0]": {'min': '12345.67890', 'max': '0', 'unit': None}
			, "[-12345.67890:0]": {'min': '-12345.67890', 'max': '0', 'unit': None}
			, "[ 1:0 ](mm)": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ]( mm)": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ]( mm )": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ](mm )": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ] ( mm )": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ] (mm )": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ] ( mm)": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ] (mm)": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ](mm) ": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ]( mm) ": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ]( mm ) ": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ](mm ) ": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ] ( mm ) ": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ] (mm ) ": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ] ( mm) ": {'min': '1', 'max': '0', 'unit': 'mm'}
			, "[ 1:0 ] (mm) ": {'min': '1', 'max': '0', 'unit': 'mm'}
		}
	}
	, "range3_schema_re":{
		"regex": openscad.range3_schema_re
		, "tests":{
			  "": None
			, " ": None
			, "[]": None
			, " [ ] ": None
			, " [] ": None
			, "[] ": None
			, " []": None
			, "[]": None
			, "[0]": None
			, "[-0]": None
			, "[1]": None
			, "[-1]": None
			, "[.0]": None
			, "[-.0]": None
			, "[-0.0]": None
			, "[-1.0]": None
			, "[12345.67890]": None
			, "[-12345.67890]": None
			, "[ 1 ](mm)": None
			, "[ 1 ]( mm)": None
			, "[ 1 ]( mm )": None
			, "[ 1 ](mm )": None
			, "[ 1 ] ( mm )": None
			, "[ 1 ] (mm )": None
			, "[ 1 ] ( mm)": None
			, "[ 1 ] (mm)": None
			, "[ 1 ](mm) ": None
			, "[ 1 ]( mm) ": None
			, "[ 1 ]( mm ) ": None
			, "[ 1 ](mm ) ": None
			, "[ 1 ] ( mm ) ": None
			, "[ 1 ] (mm ) ": None
			, "[ 1 ] ( mm) ": None
			, "[ 1 ] (mm) ": None
			, "[0:0]": None
			, "[-0:0]": None
			, "[1:0]": None
			, "[-1:0]": None
			, "[.0:0]": None
			, "[-.0:0]": None
			, "[-0.0:0]": None
			, "[-1.0:0]": None
			, "[12345.67890:0]": None
			, "[-12345.67890:0]": None
			, "[ 1:0 ](mm)": None
			, "[ 1:0 ]( mm)": None
			, "[ 1:0 ]( mm )": None
			, "[ 1:0 ](mm )": None
			, "[ 1:0 ] ( mm )": None
			, "[ 1:0 ] (mm )": None
			, "[ 1:0 ] ( mm)": None
			, "[ 1:0 ] (mm)": None
			, "[ 1:0 ](mm) ": None
			, "[ 1:0 ]( mm) ": None
			, "[ 1:0 ]( mm ) ": None
			, "[ 1:0 ](mm ) ": None
			, "[ 1:0 ] ( mm ) ": None
			, "[ 1:0 ] (mm ) ": None
			, "[ 1:0 ] ( mm) ": None
			, "[ 1:0 ] (mm) ": None
			, "[0:0:0]": {'min': '0', 'step': '0', 'max': '0', 'unit': None}
			, "[-0:0:0]": {'min': '-0', 'step': '0', 'max': '0', 'unit': None}
			, "[1:0:0]": {'min': '1', 'step': '0', 'max': '0', 'unit': None}
			, "[-1:0:0]": {'min': '-1', 'step': '0', 'max': '0', 'unit': None}
			, "[.0:0:0]": {'min': '.0', 'step': '0', 'max': '0', 'unit': None}
			, "[-.0:0:0]": {'min': '-.0', 'step': '0', 'max': '0', 'unit': None}
			, "[-0.0:0:0]": {'min': '-0.0', 'step': '0', 'max': '0', 'unit': None}
			, "[-1.0:0:0]": {'min': '-1.0', 'step': '0', 'max': '0', 'unit': None}
			, "[12345.67890:0:0]": {'min': '12345.67890', 'step': '0', 'max': '0', 'unit': None}
			, "[-12345.67890:0:0]": {'min': '-12345.67890', 'step': '0', 'max': '0', 'unit': None}
			, "[ 1:0:0 ](mm)": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ]( mm)": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ]( mm )": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ](mm )": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ] ( mm )": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ] (mm )": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ] ( mm)": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ] (mm)": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ](mm) ": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ]( mm) ": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ]( mm ) ": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ](mm ) ": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ] ( mm ) ": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ] (mm ) ": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ] ( mm) ": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
			, "[ 1:0:0 ] (mm) ": {'min': '1', 'step': '0', 'max': '0', 'unit': 'mm'}
		}
	}
	, "labeled_number_schema_re":{
		"regex": openscad.labeled_number_schema_re
		, "tests":{
			  "": None
			, " ": None
			, "[]": None
			, " [ ] ": None
			, " [] ": None
			, "[] ": None
			, " []": None
			, "[]": None
			, "[ value1, value2, value3 ]": None
			, "[value1,value2,value3]": None
			, "[value1, value2, value3 ]": None
			, "[ value1 , value2 , value3 ]": None
			, "[value1:, value2, value3 ]": None
			, "[ value1: , value2 , value3 ]": None
			, "[ 1, 2, 3 ]": None
			, "[1,2,3]": None
			, "[1, 2, 3 ]": None
			, "[ 1 , 2 , 3 ]": None
			, "[1:, 2, 3 ]": None
			, "[ 1: , 2 , 3 ]": None

			, "[value1 :label1,value2: label2,value3:label3]": None
			, "[value1 :label1, value2: label2, value3:label3]": None
			, "[value1 :label1 ,value2: label2 ,value3:label3]": None
			, "[value1 :label1 , value2: label2 , value3:label3]": None
			, "[ value1 :label1,value2: label2,value3:label3]": None
			, "[ value1 :label1 ,value2: label2 ,value3:label3]": None
			, "[ value1 :label1, value2: label2, value3:label3]": None
			, "[ value1 :label1 , value2: label2 , value3:label3]": None
			, "[value1 :label1,value2: label2,value3:label3 ]": None
			, "[value1 :label1, value2: label2, value3:label3 ]": None
			, "[value1 :label1 ,value2: label2 ,value3:label3 ]": None
			, "[value1 :label1 , value2: label2 , value3:label3 ]": None
			, "[ value1 :label1,value2: label2,value3:label3 ]": None
			, "[ value1 :label1, value2: label2, value3:label3 ]": None
			, "[ value1 :label1 ,value2: label2 ,value3:label3 ]": None
			, "[ value1 :label1 , value2: label2 , value3:label3 ]": None
			
			
			, "[1 :label1,2: label2,3:label3]": {'values': '1 :label1,2: label2,3:label3', 'unit': None}
			, "[1 :label1, 2: label2, 3:label3]": {'values': '1 :label1, 2: label2, 3:label3', 'unit': None}
			, "[1 :label1 ,2: label2 ,3:label3]": {'values': '1 :label1 ,2: label2 ,3:label3', 'unit': None}
			, "[1 :label1 , 2: label2 , 3:label3]": {'values': '1 :label1 , 2: label2 , 3:label3', 'unit': None}
			, "[ 1 :label1,2: label2,3:label3]": {'values': '1 :label1,2: label2,3:label3', 'unit': None}
			, "[ 1 :label1 ,2: label2 ,3:label3]": {'values': '1 :label1 ,2: label2 ,3:label3', 'unit': None}
			, "[ 1 :label1, 2: label2, 3:label3]": {'values': '1 :label1, 2: label2, 3:label3', 'unit': None}
			, "[ 1 :label1 , 2: label2 , 3:label3]": {'values': '1 :label1 , 2: label2 , 3:label3', 'unit': None}
			, "[1 :label1,2: label2,3:label3 ]": {'values': '1 :label1,2: label2,3:label3', 'unit': None}
			, "[1 :label1 ,2: label2 ,3:label3 ]": {'values': '1 :label1 ,2: label2 ,3:label3', 'unit': None}
			, "[1 :label1, 2: label2, 3:label3 ]": {'values': '1 :label1, 2: label2, 3:label3', 'unit': None}
			, "[1 :label1 , 2: label2 , 3:label3 ]": {'values': '1 :label1 , 2: label2 , 3:label3', 'unit': None}
			, "[ 1 :label1,2: label2,3:label3 ]": {'values': '1 :label1,2: label2,3:label3', 'unit': None}
			, "[ 1 :label1, 2: label2, 3:label3 ]": {'values': '1 :label1, 2: label2, 3:label3', 'unit': None}
			, "[ 1 :label1 ,2: label2 ,3:label3 ]": {'values': '1 :label1 ,2: label2 ,3:label3', 'unit': None}
			, "[ 1 :label1 , 2: label2 , 3:label3 ]": {'values': '1 :label1 , 2: label2 , 3:label3', 'unit': None}
			
			, "[ 1:label1 ](mm) ": {'unit': 'mm', 'values': '1:label1'}
			, "[ 1:label1 ]( mm) ": {'unit': 'mm', 'values': '1:label1'}
			, "[ 1:label1 ]( mm ) ": {'unit': 'mm', 'values': '1:label1'}
			, "[ 1:label1 ](mm ) ": {'unit': 'mm', 'values': '1:label1'}
			, "[ 1:label1 ] ( mm ) ": {'unit': 'mm', 'values': '1:label1'}
			, "[ 1:label1 ] (mm ) ": {'unit': 'mm', 'values': '1:label1'}
			, "[ 1:label1 ] ( mm) ": {'unit': 'mm', 'values': '1:label1'}
			, "[ 1:label1 ] (mm) ": {'unit': 'mm', 'values': '1:label1'}

		}
	}
	, "values_number_schema_re":{
		"regex": openscad.values_number_schema_re
		, "tests":{
			  "": None
			, " ": None
			, "[]": None
			, " [ ] ": None
			, " [] ": None
			, "[] ": None
			, " []": None
			, "[]": None
			, "[ value1, value2, value3 ]": None
			, "[value1,value2,value3]": None
			, "[value1, value2, value3 ]": None
			, "[ value1 , value2 , value3 ]": None
			, "[value1:, value2, value3 ]": None
			, "[ value1: , value2 , value3 ]": None
			, "[1,2,3]": {'values': '1,2,3', 'unit': None}
			, "[ 1,2,3]": {'values': '1,2,3', 'unit': None}
			, "[1,2,3 ]": {'values': '1,2,3', 'unit': None}
			, "[ 1,2,3 ]": {'values': '1,2,3', 'unit': None}
			, "[ 1, 2 ,3 ]": {'values': '1, 2 ,3', 'unit': None}
			, "[1, 2 ,3]": {'values': '1, 2 ,3', 'unit': None}
			, "[1.0,2.0,3.0]": {'values': '1.0,2.0,3.0', 'unit': None}
			, "[ 1.0,2.0,3.0]": {'values': '1.0,2.0,3.0', 'unit': None}
			, "[1.0,2.0,3.0 ]": {'values': '1.0,2.0,3.0', 'unit': None}
			, "[ 1.0,2.0,3.0 ]": {'values': '1.0,2.0,3.0', 'unit': None}
			, "[ 1.0, 2.0 ,3.0 ]": {'values': '1.0, 2.0 ,3.0', 'unit': None}
			, "[1.0, 2.0 ,3.0]": {'values': '1.0, 2.0 ,3.0', 'unit': None}
			, "[-1.0,-2.0,-3.0]": {'values': '-1.0,-2.0,-3.0', 'unit': None}
			, "[ -1.0,-2.0,-3.0]": {'values': '-1.0,-2.0,-3.0', 'unit': None}
			, "[-1.0,-2.0,-3.0 ]": {'values': '-1.0,-2.0,-3.0', 'unit': None}
			, "[ -1.0,-2.0,-3.0 ]": {'values': '-1.0,-2.0,-3.0', 'unit': None}
			, "[ -1.0, -2.0 ,-3.0 ]": {'values': '-1.0, -2.0 ,-3.0', 'unit': None}
			, "[-1.0, -2.0 ,-3.0]": {'values': '-1.0, -2.0 ,-3.0', 'unit': None}
			, "[-1.0, -2.0 ,-3.0,-1.0, -2.0 ,-3.0]": {'values': '-1.0, -2.0 ,-3.0,-1.0, -2.0 ,-3.0', 'unit': None}
			, "[1,2,3](mm)": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3]( mm)": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3]( mm )": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3](mm )": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3] ( mm )": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3] (mm )": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3] ( mm)": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3] (mm)": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3](mm) ": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3]( mm) ": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3]( mm ) ": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3](mm ) ": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3] ( mm ) ": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3] (mm ) ": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3] ( mm) ": {'values': '1,2,3', 'unit': 'mm'}
			, "[1,2,3] (mm) ": {'values': '1,2,3', 'unit': 'mm'}
			, "[ 1,2,3] ( mm ) ": {'values': '1,2,3', 'unit': 'mm'}
			, "[ 1.0,2.0,3.0 ] ( mm ) ": {'values': '1.0,2.0,3.0', 'unit': 'mm'}
			, "[ 1.0, 2.0 ,3.0 ] ( mm ) ": {'values': '1.0, 2.0 ,3.0', 'unit': 'mm'}
			, "[1.0, 2.0 ,3.0] ( mm ) ": {'values': '1.0, 2.0 ,3.0', 'unit': 'mm'}
			, "[-1.0,-2.0,-3.0] ( mm ) ": {'values': '-1.0,-2.0,-3.0', 'unit': 'mm'}
			, "[ -1.0, -2.0 ,-3.0 ] ( mm ) ": {'values': '-1.0, -2.0 ,-3.0', 'unit': 'mm'}
			, "[-1.0, -2.0 ,-3.0] ( mm ) ": {'values': '-1.0, -2.0 ,-3.0', 'unit': 'mm'}
			, "[-1.0, -2.0 ,-3.0,-1.0, -2.0 ,-3.0] ( mm ) ": {'values': '-1.0, -2.0 ,-3.0,-1.0, -2.0 ,-3.0', 'unit': 'mm'}
		}
	}
	
	
	
}


def test_regexes():
	for name, case in regexes.items():
		tests = case.get("tests")
		regex = case.get("regex")
		if tests and regex:
			logger.info(f"----------------------------------------------------- {name} ({regex})")
			for test, expected in tests.items():
				m = regex.match(test)
				actual = m and m.groupdict()
				logger.info(f" + '{test}': {pprint.pformat(expected)} vs. {pprint.pformat(actual)}")
				assert actual == expected, f"For test {name}: '{test}'"
		else:
			logger.warn(f"Problem in regex '{name}' ({regex}) or it's tests '{pprint.pformat(tests)}', skipping...")
			

def test_one():
	source = """
/* [Drop down box] */
// labeled combo box for numbers
Labeled_values=10; // [10:L, 20:M, 30:XL]
"""
	expected_schema = [{'default': {'type': 'number', 'value': 10},
	'description': 'labeled combo box for numbers',
	'name': 'Labeled_values',
	'schema': {'labels': {10: 'L', 20: 'M', 30: 'XL'},
			   'raw': '[10:L, 20:M, 30:XL]'},
	'section': 'Drop down box',
	'unit':None
	}]
	schema = openscad.parse_customize(source, do_debug = False)
	if expected_schema:
		logger.info("ONE -------------------------------------------------\n"+source+"\n")
		logger.info("SCHEMA:\n\n"+pprint.pformat(schema)+"\n")
		logger.info("EXPECTED:\n\n"+pprint.pformat(expected_schema)+"\n")
		assert(expected_schema == schema)
	



